﻿
#include "Manifolds/Multinomial/Multinomial.h"

/*Define the namespace*/
namespace ROPTLIB{

	Multinomial::Multinomial(integer inn, integer inm)
	{
		n = inn;
		m = inm;
		IsIntrApproach = false;
		HasHHR = false;
		UpdBetaAlone = false;
		HasLockCon = false;
		name.assign("Multinomial Manifold");
		IntrinsicDim = n * m - n;
		ExtrinsicDim = n * m;
		EMPTYEXTR = new MNVector(n, m);
		EMPTYINTR = new MNVector(IntrinsicDim, 1);
	};

	Multinomial::~Multinomial(void)
	{
		delete EMPTYEXTR;
		delete EMPTYINTR;
	};

	void Multinomial::CheckParams(void) const
	{
		Manifold::CheckParams();
		printf("%s PARAMETERS:\n", name.c_str());
		printf("row           :%15d,\t", n);
		printf("col           :%15d\n", m);
	};

	realdp Multinomial::Metric(Variable *x, Vector *etax, Vector *xix) const
	{
		realdp result = 0;
		const realdp *xptr = x->ObtainReadData();
		const realdp *etaxptr = etax->ObtainReadData();
		const realdp *xixptr = xix->ObtainReadData();
		for (integer i = 0; i < n * m; i++)
		{
			if(xptr[i] > 0)
				result += etaxptr[i] * xixptr[i] / xptr[i];
			//std::cout << "i:" << i << ",:" << etaxptr[i] << ",:" << xixptr[i] << ",:" << xptr[i] << ", partial:" << result << std::endl;//---
		}
		return result;
	};

	void Multinomial::ExtrProjection(Variable *x, Vector *etax, Vector *result) const
	{
		etax->CopyTo(result);
		realdp *resultTV = result->ObtainWritePartialData();
		const realdp *xptr = x->ObtainReadData();
		realdp *onesm = new realdp[m + n + m * n];
		realdp *tmpv = onesm + m;
		realdp *tmp = tmpv + n;
		for (integer i = 0; i < m; i++)
			onesm[i] = 1;

		gemv_(GLOBAL::N, &n, &m, &GLOBAL::DONE, resultTV, &n, onesm, &GLOBAL::IONE, &GLOBAL::DZERO, tmpv, &GLOBAL::IONE);
		gemm_(GLOBAL::N, GLOBAL::N, &n, &m, &GLOBAL::IONE, &GLOBAL::DONE, tmpv, &n, onesm, &GLOBAL::IONE, &GLOBAL::DZERO, tmp, &n);
		for (integer i = 0; i < n * m; i++)
			tmp[i] *= xptr[i];
		
		integer length = n * m;
		axpy_(&length, &GLOBAL::DNONE, tmp, &GLOBAL::IONE, resultTV, &GLOBAL::IONE);
		delete[] onesm;
	};

	void Multinomial::EucGradToGrad(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const
	{
		if (prob->GetUseHess())
		{
			Vector *segf = egf->ConstructEmpty();
			segf->NewMemoryOnWrite(); // I don't remember the reason. It seems to be required.
			egf->CopyTo(segf);
			SharedSpace *Sharedegf = new SharedSpace(segf);
			x->AddToTempData("EGrad", Sharedegf);
		}
		const realdp *xptr = x->ObtainReadData();
		egf->CopyTo(gf);
		realdp *gfTV = gf->ObtainWriteEntireData();
		for (integer i = 0; i < n * m; i++)
			gfTV[i] *= xptr[i];

		ExtrProjection(x, gf, gf);
	};

	void Multinomial::EucHvToHv(Variable *x, Vector *etax, Vector *exix, Vector* xix, const Problem *prob) const
	{
		const realdp *xptr = x->ObtainReadData();
		const realdp *etaxptr = etax->ObtainReadData();
		const SharedSpace *Sharedegf = x->ObtainReadTempData("EGrad");
		Vector *segf = Sharedegf->GetSharedElement();
		const realdp *egf = segf->ObtainReadData();

		exix->CopyTo(xix);
		realdp *xixptr = xix->ObtainWriteEntireData();
		//segf->Print("egf:");//---
		//xix->Print("xix1:");//---
		for (integer i = 0; i < n * m; i++)
		{
			xixptr[i] *= xptr[i];
			xixptr[i] += egf[i] * etaxptr[i];
		}
		realdp *tmp = new realdp[n * m];

		for (integer i = 0; i < n * m; i++)
			tmp[i] = egf[i] * xptr[i];

		for (integer i = 0; i < n; i++)
		{
			realdp ss = 0;
			for (integer j = 0; j < m; j++)
			{
				ss += tmp[i + j * n];
			}
			for (integer j = 0; j < m; j++)
				tmp[i + j * n] = ss * etaxptr[i + j * n];
		}
		//xix->Print("xix2:");//---
		//ForDebug::Print("tmp:", tmp, n, m);//---
		integer length = n * m;
		axpy_(&length, &GLOBAL::DNONE, tmp, &GLOBAL::IONE, xixptr, &GLOBAL::IONE);

		for (integer i = 0; i < n * m; i++)
			tmp[i] = -0.5 * etaxptr[i] * egf[i] / xptr[i];
		//ForDebug::Print("tmp:", tmp, n, m);//---

		axpy_(&length, &GLOBAL::DONE, tmp, &GLOBAL::IONE, xixptr, &GLOBAL::IONE);

		delete[] tmp;
		ExtrProjection(x, xix, xix);
	};

	void Multinomial::Retraction(Variable *x, Vector *etax, Variable *result, realdp stepsize) const
	{
		const realdp *xptr = x->ObtainReadData();
		const realdp *etaxptr = etax->ObtainReadData();
		realdp *resultptr = result->ObtainWriteEntireData();
		for (integer i = 0; i < n * m; i++)
		{
			if (xptr[i] != 0)
				resultptr[i] = xptr[i] * std::exp(etaxptr[i] / xptr[i]);
			else
				resultptr[i] = 0;
		}
		for (integer i = 0; i < n; i++)
		{
			realdp ss = 0;
			for (integer j = 0; j < m; j++)
				ss += resultptr[i + j * n];
			for (integer j = 0; j < m; j++)
				resultptr[i + j * n] /= ss;
		}
	};

	void Multinomial::DiffRetraction(Variable *x, Vector *etax, Variable *y, Vector *xix, Vector *result, bool IsEtaXiSameDir) const
	{
		printf("Multinomial::DiffRetraction has not been done!\n");
		xix->CopyTo(result);

		if (IsEtaXiSameDir && (HasHHR || UpdBetaAlone))
		{
			const realdp *etaxTV = etax->ObtainReadData();
			const realdp *xixTV = xix->ObtainReadData();
			realdp EtatoXi = sqrt(Metric(x, etax, etax) / Metric(x, xix, xix));
			SharedSpace *beta = new SharedSpace(1, 3);
			realdp *betav = beta->ObtainWriteEntireData();
			betav[0] = sqrt(Metric(x, etax, etax) / Metric(x, result, result)) / EtatoXi;
			betav[1] = Metric(x, etax, etax);
			betav[2] = Metric(x, result, result) * EtatoXi * EtatoXi;
			etax->AddToTempData("beta", beta);

			if (HasHHR)
			{
				Vector *TReta = result->ConstructEmpty();
				result->CopyTo(TReta);
				ScaleTimesVector(x, betav[0] * EtatoXi, TReta, TReta);
				SharedSpace *SharedTReta = new SharedSpace(TReta);
				etax->AddToTempData("betaTReta", SharedTReta);
			}
		}
	};

	void Multinomial::coTangentVector(Variable *x, Vector *etax, Variable *y, Vector *xiy, Vector *result) const
	{
		printf("Multinomial::coTangentVector has not been done!\n");
		xiy->CopyTo(result);
	};

	realdp Multinomial::Beta(Variable *x, Vector *etax) const
	{
		if (!HasHHR && !UpdBetaAlone)
			return 1;

		if (!etax->TempDataExist("beta"))
		{ /*In case that beta is not computed, then compute it.*/
			Variable *y = x->ConstructEmpty();
			Vector *xiy = etax->ConstructEmpty();
			Retraction(x, etax, y, 1);
			DiffRetraction(x, etax, y, etax, xiy, true);
			delete y;
			delete xiy;
		}

		/*If the beta has been computed in differentiated retraction, then obtain it.
		Beta should be almost always computed before.*/
		const SharedSpace *beta = etax->ObtainReadTempData("beta");
		const realdp *betav = beta->ObtainReadData();
		return betav[0];
	};
}; /*end of ROPTLIB namespace*/
