﻿
#include "Manifolds/SPDManifold/SPDManifold.h"

/*Define the namespace*/
namespace ROPTLIB{

	SPDManifold::SPDManifold(integer inn)
	{
		n = inn;
		IsIntrApproach = true;
		HasHHR = false;
		UpdBetaAlone = false;
		HasLockCon = false;
		name.assign("SPDManifold");
		IntrinsicDim = n * (n + 1) / 2;
		ExtrinsicDim = n * n;
		metric = SPDAFFINEINVARIANCE;
		EMPTYEXTR = new SPDVector(n, n);
		EMPTYINTR = new SPDVector(IntrinsicDim, 1);
	};

	SPDManifold::~SPDManifold(void)
	{
		delete EMPTYEXTR;
		delete EMPTYINTR;
	};

	void SPDManifold::ChooseSPDParamsSet1(void)
	{
		metric = SPDAFFINEINVARIANCE;
	};

	void SPDManifold::ChooseSPDParamsSet2(void)
	{
		metric = SPDEUCLIDEAN;
	}

	void SPDManifold::CheckParams(void) const
	{
		std::string SPDMetricnames[SPDMETRICLENGTH] = { "EUCLIDEAN", "AFFINEINVARIANCE" };
		Manifold::CheckParams();
		printf("%s PARAMETERS:\n", name.c_str());
		printf("row           :%15d,\t", n);
		printf("col           :%15d\n", n);
		printf("metric        :%15s,\n", SPDMetricnames[metric].c_str());
	};

	void SPDManifold::CholeskyRepresentation(Variable *x) const
	{
		const realdp *xM = x->ObtainReadData();
		Variable *L = x->ConstructEmpty();
		SharedSpace *SharedL = new SharedSpace(L);
		realdp *LM = L->ObtainWriteEntireData();
		for (integer i = 0; i < n; i++)
		{
			for (integer j = i; j < n; j++)
			{
				LM[i + j * n] = 0;
				LM[j + i * n] = xM[j + i * n];
			}
		}

		integer info, N = n;
		potrf_(GLOBAL::L, &N, LM, &N, &info);
		x->AddToTempData("L", SharedL);
		if (info != 0)
		{
			printf("Warning: SPDManifold::CholeskyRepresentation fails with info:%d!\n", info);
		}
	};

	void SPDManifold::ExtrProjection(Variable *x, Vector *etax, Vector *result) const
	{
		const realdp *etaxTV = etax->ObtainReadData();
		realdp *resultTV = result->ObtainWriteEntireData();
		for (integer i = 0; i < n; i++)
		{
			resultTV[i + i * n] = etaxTV[i + i * n];
			for (integer j = i + 1; j < n; j++)
			{
				resultTV[i + j * n] = (etaxTV[i + j * n] + etaxTV[j + i * n]) * static_cast<realdp> (0.5);
				resultTV[j + i * n] = resultTV[i + j * n];
			}
		}
	};

	void SPDManifold::EucGradToGrad(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const
	{
		if (metric == SPDAFFINEINVARIANCE)
			EucGradToGradAF(x, egf, gf, prob);
		else
		if (metric == SPDEUCLIDEAN)
			EucGradToGradEuc(x, egf, gf, prob);
		else
		{
			printf("Warning: the Riemannian metric is not valid! Use the affine invariance metric instead!");
			EucGradToGradAF(x, egf, gf, prob);
		}
	};

	void SPDManifold::EucHvToHv(Variable *x, Vector *etax, Vector *exix, Vector* xix, const Problem *prob) const
	{
		if (metric == SPDAFFINEINVARIANCE)
			EucHvToHvAF(x, etax, exix, xix, prob);
		else
			if (metric == SPDEUCLIDEAN)
				EucHvToHvEuc(x, etax, exix, xix, prob);
			else
			{
				printf("Warning: the Riemannian metric is not valid! Use the affine invariance metric instead!");
				EucHvToHvAF(x, etax, exix, xix, prob);
			}
	};

	void SPDManifold::ObtainIntr(Variable *x, Vector *etax, Vector *result) const
	{
		if (metric == SPDAFFINEINVARIANCE)
			ObtainIntrAF(x, etax, result);
		else
			if (metric == SPDEUCLIDEAN)
				ObtainIntrEuc(x, etax, result);
			else
			{
				printf("Warning: the Riemannian metric is not valid! Use the affine invariance metric instead!");
				ObtainIntrAF(x, etax, result);
			}
	};

	void SPDManifold::ObtainExtr(Variable *x, Vector *intretax, Vector *result) const
	{
		if (metric == SPDAFFINEINVARIANCE)
			ObtainExtrAF(x, intretax, result);
		else
			if (metric == SPDEUCLIDEAN)
				ObtainExtrEuc(x, intretax, result);
			else
			{
				printf("Warning: the Riemannian metric is not valid! Use the affine invariance metric instead!");
				ObtainExtrAF(x, intretax, result);
			}
	};

	realdp SPDManifold::Dist(Variable *x1, Variable *x2) const
	{
		if (metric == SPDAFFINEINVARIANCE)
			return DistAF(x1, x2);
		else
			if (metric == SPDEUCLIDEAN)
				return DistEuc(x1, x2);
			else
			{
				printf("Warning: the Riemannian metric is not valid! Use the affine invariance metric instead!");
				return DistAF(x1, x2);
			}
	};

	void SPDManifold::Retraction(Variable *x, Vector *etax, Variable *result, realdp stepsize) const
	{
		if (!x->TempDataExist("L"))
		{
			CholeskyRepresentation(x);
		}
		const SharedSpace *SharedL = x->ObtainReadTempData("L");
		Variable *LElement = SharedL->GetSharedElement();
		const realdp *L = LElement->ObtainReadData();

		/*Compute the extrinsic representation*/
		Vector *exetax = nullptr;
		const realdp *etaxTV = nullptr;
		if (IsIntrApproach)
		{
			exetax = EMPTYEXTR->ConstructEmpty();
			ObtainExtr(x, etax, exetax);
			etaxTV = exetax->ObtainReadData();
		}
		else
		{
			etaxTV = etax->ObtainReadData();
		}

		integer N = n, info;
		realdp *LiE = new realdp[N * N];
		integer length = n * n;
		copy_(&length, const_cast<realdp *> (etaxTV), &GLOBAL::IONE, LiE, &GLOBAL::IONE);
		/*Solve the linear system L X = E, i.e., X = L^{-1} E. The solution X is stored in LiE
		Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
		trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, const_cast<realdp *> (L), &N, LiE, &N, &info);
		if (info != 0)
		{
			printf("warning: SPDManifold::Retraction fails with info:%d!\n", info);
		}
		realdp *resultTV = result->ObtainWriteEntireData();
		/*Compute result = LiE^T LiE = E L^{-T} L^{-1} E*/
		gemm_(GLOBAL::T, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, LiE, &N, LiE, &N, &GLOBAL::DZERO, resultTV, &N);
		delete[] LiE;
		realdp scalar = 0.5;
		/*result <-- 0.5 * result*/
		scal_(&length, &scalar, resultTV, &GLOBAL::IONE);

		/*result <-- etax + result*/
		axpy_(&length, &GLOBAL::DONE, const_cast<realdp *> (etaxTV), &GLOBAL::IONE, resultTV, &GLOBAL::IONE);

		const realdp *xM = x->ObtainReadData();
		/*result <-- x + result*/
		axpy_(&length, &GLOBAL::DONE, const_cast<realdp *> (xM), &GLOBAL::IONE, resultTV, &GLOBAL::IONE);
		delete exetax;

		if (!result->TempDataExist("L"))
		{
			CholeskyRepresentation(result);
		}
	};

	void SPDManifold::DiffRetraction(Variable *x, Vector *etax, Variable *y, Vector *xix, Vector *result, bool IsEtaXiSameDir) const
	{ // xix + 0.5 xix x^{-1} etax + 0.5 etax x^{-1} xix
		if (!x->TempDataExist("L"))
		{
			CholeskyRepresentation(x);
		}
		const SharedSpace *SharedL = x->ObtainReadTempData("L");
		Variable *LElement = SharedL->GetSharedElement();
		const realdp *L = LElement->ObtainReadData();

		/*Compute the extrinsic representation*/
		Vector *exetax = EMPTYEXTR->ConstructEmpty();
		Vector *exxix = EMPTYEXTR->ConstructEmpty();
		ObtainExtr(x, etax, exetax);
		ObtainExtr(x, xix, exxix);
		realdp *LiE = new realdp[2 * n * n];
		realdp *LiX = LiE + n * n;
		const realdp *exetaxTV = exetax->ObtainReadData();
		const realdp *exxixTV = exxix->ObtainReadData();
		integer length = n * n, N = n, info;
		copy_(&length, const_cast<realdp*> (exetaxTV), &GLOBAL::IONE, LiE, &GLOBAL::IONE);
		copy_(&length, const_cast<realdp*> (exxixTV), &GLOBAL::IONE, LiX, &GLOBAL::IONE);
		delete exetax;

		/*Solve the linear system L X = Eta, i.e., X = L^{-1} Eta. The solution X is stored in LiE
		Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
		trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, const_cast<realdp *> (L), &N, LiE, &N, &info);

		/*Solve the linear system L X = Xix, i.e., X = L^{-1} Xix. The solution X is stored in LiX
		Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
		trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, const_cast<realdp *> (L), &N, LiX, &N, &info);

		Vector *exresult = EMPTYEXTR->ConstructEmpty();
		realdp *resultTV = exresult->ObtainWriteEntireData();
		/*resultTV <-- etax L^{-T} L^{-1} xix*/
		gemm_(GLOBAL::T, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, LiE, &N, LiX, &N, &GLOBAL::DZERO, resultTV, &N);
		for (integer i = 0; i < n; i++)
		{
			for (integer j = i + 1; j < n; j++)
			{
				resultTV[j + i * n] = (resultTV[j + i * n] + resultTV[i + j * n]) * static_cast<realdp> (0.5);
				resultTV[i + j * n] = resultTV[j + i * n];
			}
		}
		delete[] LiE;

		axpy_(&length, &GLOBAL::DONE, const_cast<realdp*> (exxixTV), &GLOBAL::IONE, resultTV, &GLOBAL::IONE);
		delete exxix;
		ObtainIntr(y, exresult, result);
		delete exresult;

		if (IsEtaXiSameDir && (HasHHR || UpdBetaAlone))
		{
			const realdp *etaxTV = etax->ObtainReadData();
			const realdp *xixTV = xix->ObtainReadData();
			realdp EtatoXi = sqrt(Metric(x, etax, etax) / Metric(x, xix, xix));
			SharedSpace *beta = new SharedSpace(1, 3);
			realdp *betav = beta->ObtainWriteEntireData();
			betav[0] = sqrt(Metric(x, etax, etax) / Metric(x, result, result)) / EtatoXi;
			betav[1] = Metric(x, etax, etax);
			betav[2] = Metric(x, result, result) * EtatoXi * EtatoXi;
			etax->AddToTempData("beta", beta);

			if (HasHHR)
			{
				Vector *TReta = result->ConstructEmpty();
				result->CopyTo(TReta);
				ScaleTimesVector(x, betav[0] * EtatoXi, TReta, TReta);
				SharedSpace *SharedTReta = new SharedSpace(TReta);
				etax->AddToTempData("betaTReta", SharedTReta);
			}
		}
	};

	void SPDManifold::coTangentVector(Variable *x, Vector *etax, Variable *y, Vector *xiy, Vector *result) const
	{
		printf("SPDManifold::coTangentVector has not been done!\n");
		xiy->CopyTo(result);
	};

	realdp SPDManifold::Beta(Variable *x, Vector *etax) const
	{
		if (!HasHHR && !UpdBetaAlone)
			return 1;

		if (!etax->TempDataExist("beta"))
		{ /*In case that beta is not computed, then compute it.*/
			Variable *y = x->ConstructEmpty();
			Vector *xiy = etax->ConstructEmpty();
			Retraction(x, etax, y, 1);
			DiffRetraction(x, etax, y, etax, xiy, true);
			delete y;
			delete xiy;
		}

		/*If the beta has been computed in differentiated retraction, then obtain it.
		Beta should be almost always computed before.*/
		const SharedSpace *beta = etax->ObtainReadTempData("beta");
		const realdp *betav = beta->ObtainReadData();
		return betav[0];
	};

	void SPDManifold::EucGradToGradAF(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const
	{
		const realdp *xptr = x->ObtainReadData();
		const realdp *egfptr = egf->ObtainReadData();
		realdp *gfptr = gf->ObtainWriteEntireData();
		integer dim = n;
		realdp *tmp = new realdp[dim * dim];
		gemm_(GLOBAL::T, GLOBAL::N, &dim, &dim, &dim, &GLOBAL::DONE, const_cast<realdp *> (xptr), &dim, const_cast<realdp *> (egfptr), &dim,
			&GLOBAL::DZERO, tmp, &dim);
		gemm_(GLOBAL::N, GLOBAL::N, &dim, &dim, &dim, &GLOBAL::DONE, const_cast<realdp *> (tmp), &dim, const_cast<realdp *> (xptr), &dim,
			&GLOBAL::DZERO, gfptr, &dim);
		delete[] tmp;
	};

	void SPDManifold::EucHvToHvAF(Variable *x, Vector *etax, Vector *exix, Vector* xix, const Problem *prob) const
	{
		printf("warning:SPDManifold::EucHvToHv has not been done!\n");
		exix->CopyTo(xix);
	};

	void SPDManifold::ObtainIntrAF(Variable *x, Vector *etax, Vector *result) const
	{//L^{-1} etax L^{-T}, where x = L L^T, etax is assumed to be a symmetric matrix.
		if (!x->TempDataExist("L"))
		{
			CholeskyRepresentation(x);
		}
		const SharedSpace *SharedL = x->ObtainReadTempData("L");
		Variable *LElement = SharedL->GetSharedElement();
		const realdp *L = LElement->ObtainReadData();
		realdp *E = new realdp[n * n];
		integer length = n * n, N = n, info;
		copy_(&length, const_cast<realdp *> (etax->ObtainReadData()), &GLOBAL::IONE, E, &GLOBAL::IONE);
		/*Solve the linear system L X = E, i.e., X = L^{-1} E. The solution X is stored in E
		Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html*/
		trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, const_cast<realdp *> (L), &N, E, &N, &info);
		if (info != 0)
		{
			printf("warning: SPDManifold::ObtainIntr fails with info:%d!\n", info);
		}

		/*E <-- E^T*/
		realdp tmp;
		for (integer i = 0; i < n; i++)
		{
			for (integer j = i + 1; j < n; j++)
			{
				tmp = E[i + j * n];
				E[i + j * n] = E[j + i * n];
				E[j + i * n] = tmp;
			}
		}
		/*Solve the linear system L X = E, i.e., X = L^{-1} E. The solution X is stored in E
		Details: http://www.netlib.org/lapack/explore-html/d6/d6f/trtrs_8f.html */
		trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, const_cast<realdp *> (L), &N, E, &N, &info);
		if (info != 0)
		{
			printf("warning: SPDManifold::ObtainIntr fails with info:%d!\n", info);
		}
		/*We don't have to do: E <-- E^T, since E is symmetric*/
		realdp *resultTV = result->ObtainWriteEntireData();
		integer idx = 0;
		realdp r2 = static_cast<realdp> (sqrt(2.0));
		for (integer i = 0; i < n; i++)
		{
			resultTV[idx] = E[i + i * n];
			idx++;
		}

		for (integer i = 0; i < n; i++)
		{
			for (integer j = i + 1; j < n; j++)
			{
				resultTV[idx] = E[j + i * n] * r2;
				idx++;
			}
		}

		delete[] E;
	};

	void SPDManifold::ObtainExtrAF(Variable *x, Vector *intretax, Vector *result) const
	{
		if (!x->TempDataExist("L"))
		{
			CholeskyRepresentation(x);
		}
		const SharedSpace *SharedL = x->ObtainReadTempData("L");
		Variable *LElement = SharedL->GetSharedElement();
		const realdp *L = LElement->ObtainReadData();
		const realdp *intretaxTV = intretax->ObtainReadData();
		realdp *resultTV = result->ObtainWriteEntireData();

		integer idx = 0;
		realdp r2 = static_cast<realdp> (sqrt(2.0));
		for (integer i = 0; i < n; i++)
		{
			resultTV[i + i * n] = intretaxTV[idx];
			idx++;
		}

		for (integer i = 0; i < n; i++)
		{
			for (integer j = i + 1; j < n; j++)
			{
				resultTV[j + i * n] = intretaxTV[idx] / r2;
				resultTV[i + j * n] = resultTV[j + i * n];
				idx++;
			}
		}

		realdp *E = new realdp[n * n];
		integer N = n;

		/*E <-- L resultTV */
		gemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, const_cast<realdp *> (L), &N, resultTV, &N,
			&GLOBAL::DZERO, E, &N);
		/*resultTV <-- E L^T */
		gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, E, &N, const_cast<realdp *> (L), &N,
			&GLOBAL::DZERO, resultTV, &N);

		delete[] E;
	};

	realdp SPDManifold::DistAF(Variable *x1, Variable *x2) const
	{
		// dis(x, y) = |log(Lx^{-1} y Lx^{-T})|_F
		if (!x1->TempDataExist("L"))
		{
			CholeskyRepresentation(x1);
		}
		const SharedSpace *SharedLx1 = x1->ObtainReadTempData("L");
		Variable *LElementx1 = SharedLx1->GetSharedElement();
		const realdp *Lx1 = LElementx1->ObtainReadData();

		if (!x2->TempDataExist("L"))
		{
			CholeskyRepresentation(x2);
		}
		const SharedSpace *SharedLx2 = x2->ObtainReadTempData("L");
		Variable *LElementx2 = SharedLx2->GetSharedElement();
		const realdp *Lx2 = LElementx2->ObtainReadData();

		integer N = n, info, length = n * n;
		realdp *LxLy = new realdp[N * N];

		copy_(&length, const_cast<realdp *> (Lx2), &GLOBAL::IONE, LxLy, &GLOBAL::IONE);
		/*Solve the linear system Lx X = Ly, i.e., X = Lx^{-1}Ly. The solution X is stored in LxLy*/
		trtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, const_cast<realdp *> (Lx1), &N, LxLy, &N, &info);

		/* compute temp = LxLy*(LxLy)^T  */
		realdp *temp = new realdp[N * N];
		gemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, LxLy, &N, LxLy, &N, &GLOBAL::DZERO, temp, &N);

		realdp *eigenvalues = new realdp[n + n * n];
		realdp *eigenvectors = eigenvalues + n;
		Matrix E(eigenvalues, n, 1), V(eigenvectors, n, n);
		Matrix MMt(temp, n, n);
		Matrix::EigenSymmetricM(GLOBAL::L, MMt, E, V);

		realdp result = 0.0;
		for (integer i = 0; i < N; i++)
		{
			result = result + log(eigenvalues[i])*log(eigenvalues[i]);
		}

		result = sqrt(result);

		delete[] eigenvalues;
		delete[] temp;
		delete[] LxLy;
		return result;
	};



	void SPDManifold::EucGradToGradEuc(Variable *x, Vector *egf, Vector *gf, const Problem *prob) const
	{
		egf->CopyTo(gf);
	};

	void SPDManifold::EucHvToHvEuc(Variable *x, Vector *etax, Vector *exix, Vector* xix, const Problem *prob) const
	{
		exix->CopyTo(xix);
	};

	void SPDManifold::ObtainIntrEuc(Variable *x, Vector *etax, Vector *result) const
	{
		const realdp *etaxTV = etax->ObtainReadData();
		realdp *resultTV = result->ObtainWriteEntireData();
		integer idx = 0;
		realdp r2 = static_cast<realdp> (sqrt(2.0));
		for (integer i = 0; i < n; i++)
		{
			resultTV[idx] = etaxTV[i + i * n];
			idx++;
		}

		for (integer i = 0; i < n; i++)
		{
			for (integer j = i + 1; j < n; j++)
			{
				resultTV[idx] = etaxTV[j + i * n] * r2;
				idx++;
			}
		}
	};

	void SPDManifold::ObtainExtrEuc(Variable *x, Vector *intretax, Vector *result) const
	{
		const realdp *intretaxTV = intretax->ObtainReadData();
		realdp *resultTV = result->ObtainWriteEntireData();

		integer idx = 0;
		realdp r2 = static_cast<realdp> (sqrt(2.0));
		for (integer i = 0; i < n; i++)
		{
			resultTV[i + i * n] = intretaxTV[idx];
			idx++;
		}

		for (integer i = 0; i < n; i++)
		{
			for (integer j = i + 1; j < n; j++)
			{
				resultTV[j + i * n] = intretaxTV[idx] / r2;
				resultTV[i + j * n] = resultTV[j + i * n];
				idx++;
			}
		}
	};

	realdp SPDManifold::DistEuc(Variable *x1, Variable *x2) const
	{
		integer N = n, length = n * n;
		realdp result = 0.0, error;
		const realdp *x1M = x1->ObtainReadData();
		const realdp *x2M = x2->ObtainReadData();

		for (integer i = 0; i < length; i++)
		{
			//             error = const_cast<realdp *>(x1M + i) - const_cast<realdp *>(x2M + i);
			error = x1M[i] - x2M[i];
			result += error * error;
		}

		result = (static_cast<realdp> (1.0) / n) * sqrt(result);
		return result;
	};
}; /*end of ROPTLIB namespace*/
