
#include "Problems/WeightedLowrank/WeightedLowRank.h"

/*Define the namespace*/
namespace ROPTLIB{

	WeightedLowRank::WeightedLowRank(realdp *inA, realdp *inW, integer inm, integer inn, integer inr)
	{
		A = inA;
		W = inW;
		m = inm;
		n = inn;
		r = inr;
	};

	WeightedLowRank::~WeightedLowRank(void)
	{
	};

	realdp WeightedLowRank::f(Variable *x) const
	{
		ProductElement *ProdxxM = dynamic_cast<ProductElement *>(x);
		const realdp *Uptr = ProdxxM->GetElement(0)->ObtainReadData();
		const realdp *Dptr = ProdxxM->GetElement(1)->ObtainReadData();
		const realdp *Vptr = ProdxxM->GetElement(2)->ObtainReadData();

		char *transn = const_cast<char *> ("n");
		char *transt = const_cast<char *> ("t");
		char *uplo = const_cast<char *> ("u");
		realdp one = 1, zero = 0, neg_one = -1;
		integer inc = 1, M = m, R = r, N = n, MN = m * n, MR = m * r;

		realdp *UDptr = new realdp[MR];
		// UDptr <- Uptr * Dptr, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(transn, transn, &M, &R, &R, &one, const_cast<realdp *> (Uptr), &M, const_cast<realdp *> (Dptr), &R, &zero, UDptr, &M);
		SharedSpace *Temp1 = new SharedSpace(2, m, n);
		realdp *Xptr = Temp1->ObtainWriteEntireData();
		// Xptr <- UDptr * Vptr^T, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
		gemm_(transn, transt, &M, &N, &R, &one, UDptr, &M, const_cast<realdp *> (Vptr), &N, &zero, Xptr, &M);
		delete[] UDptr;

		SharedSpace *Temp2 = new SharedSpace(2, m, n);
		realdp *Errptr = Temp2->ObtainWriteEntireData();
		// Errptr <- A, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
		copy_(&MN, A, &inc, Errptr, &inc);
		// Errptr <- Errptr - Xptr, details: http://www.netlib.org/lapack/explore-html/d9/dcd/axpy_8f.html
		axpy_(&MN, &neg_one, Xptr, &inc, Errptr, &inc);

		SharedSpace *Temp3 = new SharedSpace(2, m, n);
		realdp *QXptr = Temp3->ObtainWriteEntireData();
		// QXptr = W * Errptr, details: http://www.netlib.org/lapack/explore-html/d8/dbe/symv_8f.html
		symv_(uplo, &MN, &one, W, &MN, Errptr, &inc, &zero, QXptr, &inc);

		realdp result = 0;
		// compute Errptr(:)^T QXptr(:), details: http://www.netlib.org/lapack/explore-html/d5/df6/dot_8f.html
		result = dot_(&MN, Errptr, &inc, QXptr, &inc);
		if (UseGrad)
		{
			x->AddToTempData("X", Temp1);
			x->AddToTempData("err", Temp2);
			x->AddToTempData("QX", Temp3);
		}
		else
		{
			delete Temp1;
			delete Temp2;
			delete Temp3;
		}
		return result;
	};
	
	void WeightedLowRank::EucGrad(Variable *x, Vector *egf) const
	{
		ProductElement *ProdxxM = dynamic_cast<ProductElement *>(x);
	    realdp *Uptr = const_cast<realdp *> (ProdxxM->GetElement(0)->ObtainReadData());
	   	realdp *Dptr = const_cast<realdp *> (ProdxxM->GetElement(1)->ObtainReadData());
	    realdp *Vptr = const_cast<realdp *> (ProdxxM->GetElement(2)->ObtainReadData());
		char *transn = const_cast<char *> ("n");
	  	char *transt = const_cast<char *> ("t");
	    realdp one = 1, zero = 0, neg_one = -1.0, neg_two = -2.0;
	    integer inc = 1, M = m, R = r, N = n, MN = m * n, MR = m * r, NR = n * r; 
	
	    const SharedSpace *Temp = x->ObtainReadTempData("QX");
		const realdp *QXptr = Temp->ObtainReadData();

		SharedSpace *EucRep = new SharedSpace(1, 1 + m * n);
		realdp *EucRepptr = EucRep->ObtainWriteEntireData();
		EucRepptr[0] = 0;
		realdp *fullgrad = EucRepptr + 1;

		copy_(&MN, const_cast<realdp *> (QXptr), &inc, fullgrad, &inc);
		scal_(&MN, &neg_two, fullgrad, &inc);

		egf->AddToTempData("EucRep", EucRep);
	};
	
	void WeightedLowRank::EucHessianEta(Variable *x, Vector *etax, Vector *exix) const
	{
		dynamic_cast<LowRank *> (Domain)->ExtrToEucRep(x, etax);
		const SharedSpace *EucRep = etax->ObtainReadTempData("EucRep");
		const realdp *EucRepptr = EucRep->ObtainReadData();

		SharedSpace *EucRep_R = new SharedSpace(1, 1 + m * n);
		realdp *EucRep_Rptr = EucRep_R->ObtainWriteEntireData();
		EucRep_Rptr[0] = 0;
		integer MN = n * m;
		// exix = 2 * W * etax, details: http://www.netlib.org/lapack/explore-html/d8/dbe/symv_8f.html
		symv_(GLOBAL::U, &MN, &GLOBAL::DTWO, W, &MN, const_cast<realdp *> (EucRepptr + 1), &GLOBAL::IONE, &GLOBAL::DZERO, EucRep_Rptr + 1, &GLOBAL::IONE);

		exix->AddToTempData("EucRep", EucRep_R);
	};

	//void WeightedLowRank::RieGrad(Variable *x, Vector *gf) const
	//{
	//	ProductElement *ProdxxM = dynamic_cast<ProductElement *>(x);
	//	const realdp *Uptr = ProdxxM->GetElement(0)->ObtainReadData();
	//	const realdp *Dptr = ProdxxM->GetElement(1)->ObtainReadData();
	//	const realdp *Vptr = ProdxxM->GetElement(2)->ObtainReadData();
	//	char *transn = const_cast<char *> ("n");
	//	char *transt = const_cast<char *> ("t");
	//	realdp one = 1, zero = 0, neg_one = -1.0, neg_two = -2.0;
	//	integer inc = 1, M = m, R = r, N = n, MN = m * n, MR = m * r, NR = n * r, RR = r * r;
	//	const SharedSpace *Temp = x->ObtainReadTempData("QX");
	//	const realdp *QXptr = Temp->ObtainReadData();
	//	realdp *fullgrad = new realdp[MN];
	//	// fullgrad <- Qxptr, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
	//	copy_(&MN, const_cast<realdp *> (QXptr), &inc, fullgrad, &inc);
	//	// fullgrad <- -2 * fullgrad, details: http://www.netlib.org/lapack/explore-html/d4/dd0/scal_8f.html
	//	scal_(&MN, &neg_two, fullgrad, &inc);
	//	realdp *XiVptr = new realdp[MR];
	//	// XiVptr <- fullgrad * Vptr, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
	//	gemm_(transn, transn, &M, &R, &N, &one, fullgrad, &M, const_cast<realdp *> (Vptr), &N, &zero, XiVptr, &M);

	//	realdp *XiUptr = new realdp[NR];
	//	// XiUptr <- fullgrad^T * Uptr, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
	//	gemm_(transt, transn, &N, &R, &M, &one, fullgrad, &M, const_cast<realdp *> (Uptr), &M, &zero, XiUptr, &N);
	//	delete[] fullgrad;

	//	// compute inverse of D
	//	integer info;
	//	integer *IPIV = new integer[R + 1];
	//	realdp *WORK = new realdp[RR];
	//	realdp *Dinv = new realdp[RR];
	//	// Dinv <- Dptr, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
	//	copy_(&RR, const_cast<realdp *> (Dptr), &inc, Dinv, &inc);
	//	// LU factorization for Dinv, LU factors are stored in Dinv,
	//	// details: http://www.netlib.org/lapack/explore-html/d3/d6a/getrf_8f.html
	//	getrf_(&R, &R, Dinv, &R, IPIV, &info);
	//	// Compute the inverse of Dinv based on Dinv's LU factorization,
	//	// details: http://www.netlib.org/lapack/explore-html/df/da4/getri_8f.html
	//	getri_(&R, Dinv, &R, IPIV, WORK, &RR, &info);
	//	delete[] IPIV;
	//	delete[] WORK;

	//	realdp *gfTV = gf->ObtainWriteEntireData();
	//	realdp *Udotptr = gfTV;
	//	realdp *Ddotptr = gfTV + m * r;
	//	realdp *Vdotptr = Ddotptr + r * r;
	//	// Ddotptr <- Uptr^T * XiVptr, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
	//	gemm_(transt, transn, &R, &R, &M, &one, const_cast<realdp *> (Uptr), &M, XiVptr, &M, &zero, Ddotptr, &R);
	//	// Udotptr <- Uptr * Ddotptr, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
	//	gemm_(transn, transn, &M, &R, &R, &one, const_cast<realdp *> (Uptr), &M, Ddotptr, &R, &zero, Udotptr, &M);
	//	// Udotptr <- -1 * Udotptr, details: http://www.netlib.org/lapack/explore-html/d4/dd0/scal_8f.html
	//	scal_(&MR, &neg_one, Udotptr, &inc);
	//	// Udotptr <- XiVptr + Udotptr, details: http://www.netlib.org/lapack/explore-html/d9/dcd/axpy_8f.html
	//	axpy_(&MR, &one, XiVptr, &inc, Udotptr, &inc);
	//	// Vdotptr <- Vptr * Ddotptr^T, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
	//	gemm_(transn, transt, &N, &R, &R, &one, const_cast<realdp *> (Vptr), &N, Ddotptr, &R, &zero, Vdotptr, &N);
	//	// Vdotptr <- -1 * Vdotptr, details: http://www.netlib.org/lapack/explore-html/d4/dd0/scal_8f.html
	//	scal_(&NR, &neg_one, Vdotptr, &inc);
	//	// Vdotptr <- XiUptr + Vdotptr, details: http://www.netlib.org/lapack/explore-html/d9/dcd/axpy_8f.html
	//	axpy_(&NR, &one, XiUptr, &inc, Vdotptr, &inc);

	//	realdp *Udottemp = new realdp[MR];
	//	realdp *Vdottemp = new realdp[NR];
	//	// Udottemp <- Udotptr * Dinv, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
	//	gemm_(transn, transn, &M, &R, &R, &one, Udotptr, &M, Dinv, &R, &zero, Udottemp, &M);
	//	// Vdottemp <- Vdotptr * Dinv^T, details: http://www.netlib.org/lapack/explore-html/d7/d2b/gemm_8f.html
	//	gemm_(transn, transt, &N, &R, &R, &one, Vdotptr, &N, Dinv, &R, &zero, Vdottemp, &N);
	//	// Udotptr <- Udottemp, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
	//	copy_(&MR, Udottemp, &inc, Udotptr, &inc);
	//	// Vdotptr <- Vdottemp, details: http://www.netlib.org/lapack/explore-html/da/d6c/copy_8f.html
	//	copy_(&NR, Vdottemp, &inc, Vdotptr, &inc);
	//	//gf->Print("gf");//---
	//	delete[] Udottemp;
	//	delete[] Vdottemp;
	//	delete[] Dinv;
	//	delete[] XiUptr;
	//	delete[] XiVptr;
	//};

	//void WeightedLowRank::RieHessianEta(Variable *x, Vector *etax, Vector *xix) const
	//{
	//	etax->CopyTo(xix);

	//	/*ProductElement *ProdxxM = dynamic_cast<ProductElement *>(x);
	//	 realdp *Uptr = const_cast<realdp *> (ProdxxM->GetElement(0)->ObtainReadData());
	//	 realdp *Dptr = const_cast<realdp *> (ProdxxM->GetElement(1)->ObtainReadData());
	//	 realdp *Vptr = const_cast<realdp *> (ProdxxM->GetElement(2)->ObtainReadData());
	//	 char *transn = const_cast<char *> ("n");
	//	 char *transt = const_cast<char *> ("t");
	//	 realdp one = 1, zero = 0, neg_one = -1.0, neg_two = -2.0;
	//	 integer inc = 1, M = m, R = r, N = n, MN = m * n, MR = m * r, NR = n * r;

	//	 const SharedSpace *Temp = x->ObtainReadTempData("QX");
	//	 const realdp *QXptr = Temp->ObtainReadData();

	//	 realdp *gfTV = etax->ObtainWriteEntireData();
	//	 realdp *Udotptr = gfTV;
	//	 realdp *Ddotptr = gfTV + m * r;
	//	 realdp *Vdotptr = Ddotptr + r * r;

	//	 realdp *UDdot = new realdp[MR];
	//	 gemm_(transn, transn, &M, &R, &R, &one, Uptr, &M, Ddotptr, &R, &zero, UDdot, &M);

	//	 realdp *Qeta = new realdp[MN];
	//	 gemm_(transn, transt, &M, &N, &R, &one, Udotptr, &M, Vptr, &N, &zero, Qeta, &M);*/






	//};

}; /*end of ROPTLIB namespace*/
