
#include "Solvers/ARPG.h"

/*Define the namespace*/
namespace ROPTLIB{

	ARPG::ARPG(const Problem *prob, const Variable *initialx, const Variable *insoln):RPG(prob, initialx, insoln)
	{
        Initialization(prob, initialx, insoln);
	};

    ARPG::~ARPG()
    {
        delete gfz1;
        delete gfz2;
        delete gfy1;
        delete gfy2;
        delete z1;
        delete z2;
        delete y1;
        delete y2;
        delete [] initDy;
    };

	void ARPG::SetProbX(const Problem *prob, const Variable *initialx, const Variable *insoln)
	{
//		RPG::SetProbX(prob, initialx, insoln);
        z1 = x1->ConstructEmpty();
        z2 = x1->ConstructEmpty();
        y1 = x1->ConstructEmpty();
        y2 = x1->ConstructEmpty();
        gfz1 = prob->GetDomain()->GetEMPTYEXTR()->ConstructEmpty();
        gfz2 = prob->GetDomain()->GetEMPTYEXTR()->ConstructEmpty();
        gfy1 = prob->GetDomain()->GetEMPTYEXTR()->ConstructEmpty();
        gfy2 = prob->GetDomain()->GetEMPTYEXTR()->ConstructEmpty();
        initDy = new realdp[dimNorVec];
        for (integer i = 0; i < dimNorVec; i++)
            initDy[i] = 0;
	};

	void ARPG::SetDefaultParams()
	{
//		RPG::SetDefaultParams();
        Variant = REGULAR;
        SGIterGap = 5;
		SolverName.assign("ARPG");
	};

    void ARPG::CheckParams(void)
    {
        RPG::CheckParams();

        char YES[] = "YES";
        char NO[] = "NO";
        char *status;

        printf("ARPG METHODS PARAMETERS:\n");
        status = (SGIterGap > 0) ? YES : NO;
        printf("SGIterGap     :%15d[%s],\n", SGIterGap, status);
    };

    void ARPG::PrintInfo(void)
    {
        RPG::PrintInfo();
    };

    void ARPG::PrintGenInfo(void)
    {
        printf("i:%d,f:%.16e,df/f:%.3e,", iter, fz2, ((fz1 - fz2) / std::fabs(fz2)));

        printf("nsubgf:%.3e,", nsubgf);

        printf("time:%.2e,", static_cast<realdp>(getTickCount() - starttime) / CLK_PS);

        printf("nf:%d,ng:%d,nR:%d,", nf, ng, nR);

        if (nV != 0)
            printf("nV(nVp):%d(%d),", nV, nVp);
        
        printf("LSstatus:%s,", LSstatusSetnames[LSstatus].c_str());
        if (newslope != 0)
        {
            printf("newslope:%.3e,", newslope);
        }
        printf("initstepsize:%.3e,stepsize:%.3e,", initiallength, stepsize);
    };

    void ARPG::Run(void)
    {
        if(Weight != nullptr)
        {
            delete Weight;
            Weight = nullptr;
        }
        Weight = new EucVector(LengthW);
        
        Variable *xTemp;
        Vector *gfTemp;
        pre_funs.clear();

        Solvers::Run();

        ChooseLinesearch();

        LSstatus = SUCCESS;
        f1 = Prob->f(x1); nf++;
        f2 = f1;
        Prob->Grad(x1, gf1); ng++;
        ngf0 = sqrt(Mani->Metric(x1, gf1, gf1));
        ngf1 = ngf0; ngf2 = ngf1;
        x1->CopyTo(y1);
        gf1->CopyTo(gfy1);
        fy1 = f1;
        x1->CopyTo(z1);
        gf1->CopyTo(gfz1);
        fz1 = f1;
        newslope = 0;
        s1 = 1;
        iter = 0;
        realdp *initialstepsizeptr = nullptr, *acceptedstepsizeptr = nullptr;
        if (Debug >= FINALRESULT)
            printf("i:%d,f:%.3e,|gf|:%.3e,\n", iter, f1, ngf1);
        
        if (Debug >= ITERRESULT)
        {
            timeSeries[iter] = static_cast<realdp>(getTickCount() - starttime) / CLK_PS;
            funSeries[iter] = f1;
            gradSeries[iter] = ngf1;
            distSeries[iter] = ((soln == nullptr) ? 0 : Mani->Dist(x1, soln));
            initialstepsizeptr = new realdp[1 + Max_Iteration];
            acceptedstepsizeptr = new realdp[1 + Max_Iteration];
            initialstepsizeptr[iter] = 0;
            acceptedstepsizeptr[iter] = 0;
        }
        
        bool isstop = IsStopped();
        /*If the intrinsic representation is used, then exeta1 and exeta2 are used to store the extrinsic representations of eta1 and eta2 respectively*/
        if (Prob->GetDomain()->GetIsIntrinsic())
        {
            exeta1 = Prob->GetDomain()->GetEMPTYEXTR()->ConstructEmpty();
            exeta2 = Prob->GetDomain()->GetEMPTYEXTR()->ConstructEmpty();
        }

        /*Start the loop*/
        while ((((! isstop) && iter < Max_Iteration) || iter < Min_Iteration) && LSstatus == SUCCESS)
        {
            if(iter % SGIterGap == 0)
            {
                ProximalMap(x1, gf1, initDx, eta1);// Obtain search direction eta1
//                eta1->Print("eta1:");//---
                initialslope = Mani->Metric(x1, gf1, eta1);
                /*Compute initial step size for the next iteration*/
                InitialStepSize();
                initiallength = stepsize;
                if (Debug >= ITERRESULT)
                    initialstepsizeptr[iter + 1] = initiallength;
                /*Start a line search algorithm. If the intrinsic representation is used for the search direction, then converting it into the extrinsic representation.*/
                if (Prob->GetDomain()->GetIsIntrinsic())
                {
                    Prob->GetDomain()->ObtainExtr(x1, eta1, exeta1);
                }
                
                /*If accurate enough, then a fixed stepsize is chosen.*/
                if (ngf1 / (ngf0 + Tolerance) < Accuracy)
                {
                    if (Finalstepsize > 0)
                        stepsize = Finalstepsize;
                    else
                        stepsize = initiallength;
                    f2 = h();
                    Prob->Grad(x2, gf2); ng++;
                }
                else
                {
                    /* Call the specified linesearch algorithm.
                    Note that in the linesearch algorithm, we need to obtain
                    accepted stepsize, eta2=stepsize*eta1, x2 = R_{x_1}(eta_2), f2 = f(x2), and gf2 = grad f(x_2) */
                    if (LineSearch_LS == INPUTFUN)
                    {
                        if (Prob->GetDomain()->GetIsIntrinsic())
                        {
                            stepsize = LinesearchInput(iter, x1, exeta1, stepsize, initialslope, Prob, this);
                        }
                        else
                        {
                            stepsize = LinesearchInput(iter, x1, eta1, stepsize, initialslope, Prob, this);
                        }
                        stepsize = (stepsize < std::numeric_limits<realdp>::epsilon()) ? initiallength : stepsize;
                        initiallength = stepsize;
                        if (!IsPureLSInput)
                        {
                            SolversLS::LinesearchArmijo();
                        }
                        else
                        {
                            f2 = h();
                            Prob->Grad(x2, gf2); ng++;
                        }
                    }
                    else
                    {
                        (this->*Linesearch)();
                        if(LSstatus == MINSTEPSIZE)
                        {
                            if(SMlambda <= 1e-10 && SMtol <= 1e-24)
                            {
                                break;
                            }
                            SMtol = ((SMtol * 0.01 > 1e-24) ? SMtol * 0.01 : 1e-24); //--std::max(SMtol * 0.01, 1e-24);
                            SMlambda = ((SMlambda * 0.1 > 1e-10) ? SMlambda * 0.1 : 1e-10); //--std::max(SMlambda * 0.1, 1e-10);
                            LSstatus = SUCCESS;
                            continue;
                        }
                    }
                }
//                std::cout << "stepsize:" << stepsize << std::endl;//---
//                x2->Print("x2:");//---
                if(Variant == ADALIPSCHITZ)
                {
                    if(stepsize == initiallength)
                    {
                        adavalue /= 1.01;
                    }
                    else
                    {
                        adavalue = ((adavalue * 1.01 < 1.0) ? adavalue * 1.01 : 1.0);//-- std::min(adavalue * 1.01, 1.0);
                    }
                }
                /*Output debug information if necessary.*/
                if (LSstatus < SUCCESS && Debug > FINALRESULT)
                {
                    printf("Linesearch fails! LSstatus:%s\n", LSstatusSetnames[LSstatus].c_str());
                }

                if (std::isnan(f2) || std::isinf(f2)) /*Stop when got a nan or inf*/
                {
                    printf("New function value is either nan or inf. Stop!\n");
                    break;
                }
                /*if f2 is smaller than the function value at new iterate z1, then safeguard takes effect and restart.*/
                if(iter != 0 && f2 < fz2)
                {
                    x2->CopyTo(y1);
                    gf2->CopyTo(gfy1);
                    fy1 = f2;
                    x2->CopyTo(z1);
                    s1 = 1;
                }
                /*update the safeguard iterate*/
                z1->CopyTo(x1);
                f1 = Prob->f(x1); nf++;
                Prob->Grad(x1, gf1); ng++;
                ngf1 = sqrt(Mani->Metric(x1, gf1, gf1));
            }

//            y1->Print("y1:");//--
//            gfy1->Print("gfy1:");//--
//            ForDebug::Print("initDy:", initDy, dimNorVec);//---
            ProximalMap(y1, gfy1, initDy, eta1);// Obtain search direction eta1
//            y1->Print("y1:");//--
//            eta1->Print("eta1 y:");//---
            Mani->Retraction(y1, eta1, z2, 1.0);
            fz2 = Prob->f(z2);
            s2 = (1.0 + std::sqrt(1.0 + 4.0 * s1 * s1)) / 2.0;
//            z2->Print("z2:");//---
//            z1->Print("z1:");//----
            Mani->InvRetraction(z2, z1, zeta);
//            zeta->Print("zeta:");//---
            Mani->ScaleTimesVector(nullptr, -(s1 - 1) / s2, zeta, zeta);
            
            Mani->Retraction(z2, zeta, y2, 1);
            fy2 = Prob->f(y2);
            Prob->Grad(y2, gfy2);
            
            /*Call the function to check whether the stopping criterion is satisfied or not.
            The default function is written in Solvers.h and Solvers.cpp*/
            isstop = IsStopped();
            
            iter++;

            if (Debug >= ITERRESULT)
            {
                /*Output information*/
                if (iter % OutputGap == 0)
                {
                    PrintGenInfo();
                    PrintInfo(); // Output information specific to Algorithms
                }
                /*Store debug information in the arrays*/
                timeSeries[iter] = static_cast<realdp>(getTickCount() - starttime) / CLK_PS;
                funSeries[iter] = fz2; gradSeries[iter] = nsubgf;
                distSeries[iter] = ((soln == nullptr) ? 0 : Mani->Dist(x2, soln));
                acceptedstepsizeptr[iter] = stepsize;
            }
            
            xTemp = z1; z1 = z2; z2 = xTemp;
            gfTemp = gfz1; gfz1 = gfz2; gfz2 = gfTemp;
            xTemp = y1; y1 = y2; y2 = xTemp;
            gfTemp = gfy1; gfy1 = gfy2; gfy2 = gfTemp;
            s1 = s2;
        }
        if (Prob->GetDomain()->GetIsIntrinsic())
        {
            delete exeta1;
            delete exeta2;
        }
        ComTime = static_cast<realdp>(getTickCount() - starttime) / CLK_PS;
        if (Debug >= ITERRESULT)
        {
            lengthSeries = iter + 1;
            SharedSpace *sharedinitialstepsize = new SharedSpace(1, lengthSeries);
            SharedSpace *sharedacceptedstepsize = new SharedSpace(1, lengthSeries);
            realdp *tmp1 = sharedinitialstepsize->ObtainWriteEntireData();
            realdp *tmp2 = sharedacceptedstepsize->ObtainWriteEntireData();
            copy_(&lengthSeries, initialstepsizeptr, &GLOBAL::IONE, tmp1, &GLOBAL::IONE);
            copy_(&lengthSeries, acceptedstepsizeptr, &GLOBAL::IONE, tmp2, &GLOBAL::IONE);

            SolverTMP->AddToTempData("initialstepsize", sharedinitialstepsize);
            SolverTMP->AddToTempData("acceptedstepsize", sharedacceptedstepsize);
            delete[] initialstepsizeptr;
            delete[] acceptedstepsizeptr;
        }
        PrintFinalInfo();
        if (nsubgf != -1)
        {
            ngf1 = nsubgf;
        }
    };

    void ARPG::PrintFinalInfo(void)
    {
        RPG::PrintFinalInfo();
    };

    void ARPG::SetParams(PARAMSMAP params)
    {
        RPG::SetParams(params);
        PARAMSMAP::iterator iter;
        for (iter = params.begin(); iter != params.end(); iter++)
        {
            if (iter->first == static_cast<std::string> ("SGIterGap"))
            {
                SGIterGap = static_cast<integer> (static_cast<integer> (iter->second));
            }
        }
    };
}; /*end of ROPTLIB namespace*/
