
#include "Solvers/LRTRSR1.h"

/*Define the namespace*/
namespace ROPTLIB{

	LRTRSR1::LRTRSR1(const Problem *prob, const Variable *initialx, const Variable *insoln)
	{
		Initialization(prob, initialx, insoln);
	};

	void LRTRSR1::SetProbX(const Problem *prob, const Variable *initialx, const Variable *insoln)
	{
		SolversTR::SetProbX(prob, initialx, insoln);
		const Vector *EMPTYETA;
		if (prob->GetDomain()->GetIsIntrinsic())
			EMPTYETA = prob->GetDomain()->GetEMPTYINTR();
		else
			EMPTYETA = prob->GetDomain()->GetEMPTYEXTR();
		s = EMPTYETA->ConstructEmpty();
		y = EMPTYETA->ConstructEmpty();

		prob->SetUseGrad(true);
		prob->SetUseHess(false);
	};

	void LRTRSR1::SetDefaultParams()
	{
		SolversTR::SetDefaultParams();
		theta = static_cast<realdp> (0.1);
		kappa = static_cast<realdp> (0.1);
		isconvex = false;
		LengthSY = 4;
		S = nullptr;
		Y = nullptr;
		YMGS = nullptr;
		inpss = 0;
		inpsy = 0;
		inpyy = 0;
		Currentlength = 0;
		beginidx = 0;
		SS = nullptr;
		SY = nullptr;
		PMGQ = nullptr;
		LU_PMGQ = nullptr;
		P = nullptr;
		gamma = 1;
		SolverName.assign("LRTRSR1");

		/*LSR solver tmp variables*/
		Psi = nullptr;
		PsiTPsi = nullptr;
		RT = nullptr;
		U = nullptr;
		D = nullptr;
		P_parallel = nullptr;
		g_parallel = nullptr;
		Psig = nullptr;
		Lambda = nullptr;
		a_j = nullptr;
		tolLocalNewton = static_cast<realdp> (1e-10);
	};

	LRTRSR1::~LRTRSR1(void)
	{
		delete s;
		delete y;
		DeleteVectors(S, LengthSY);
		DeleteVectors(Y, LengthSY);
		DeleteVectors(YMGS, LengthSY);
		if (SS != nullptr)
			delete[] SS;
		if (SY != nullptr)
			delete[] SY;
		if (PMGQ != nullptr)
			delete[] PMGQ;
		if (LU_PMGQ != nullptr)
			delete[] LU_PMGQ;
		if (P != nullptr)
			delete[] P;
		if (Psi != nullptr)
			delete[] Psi;
	};

	void LRTRSR1::Run(void)
	{
		DeleteVectors(S, LengthSY);
		NewVectors(S, LengthSY);
		DeleteVectors(Y, LengthSY);
		NewVectors(Y, LengthSY);
		DeleteVectors(YMGS, LengthSY);
		NewVectors(YMGS, LengthSY);
		if (SS != nullptr)
			delete[] SS;
		SS = new realdp[LengthSY * LengthSY];
		if (SY != nullptr)
			delete[] SY;
		SY = new realdp[LengthSY * LengthSY];
		if (PMGQ != nullptr)
			delete[] PMGQ;
		PMGQ = new realdp[LengthSY * LengthSY];
		if (LU_PMGQ != nullptr)
			delete[] LU_PMGQ;
		LU_PMGQ = new realdp[LengthSY * LengthSY];
		if (P != nullptr)
			delete[] P;
		P = new integer[LengthSY];

		if (Psi != nullptr)
			delete[] Psi;
		Psi = new realdp[2 * gf1->Getlength() * LengthSY + 3 * LengthSY * LengthSY + 5 * LengthSY + 2];
		PsiTPsi = Psi + gf1->Getlength() * LengthSY;
		RT = PsiTPsi + LengthSY * LengthSY;
		U = RT + LengthSY * LengthSY;
		D = U + LengthSY * LengthSY;
		P_parallel = D + LengthSY;
		g_parallel = P_parallel + gf1->Getlength() * LengthSY;
		Psig = g_parallel + LengthSY;
		Lambda = Psig + LengthSY;
		a_j = Lambda + LengthSY + 1;

		//realdp *Q = new realdp[2 * n * k + 2 * k * k + 3 * k];
		//realdp *RT = Q + n * k;
		//realdp *U = RT + k * k;
		//realdp *D = U + k * k;
		//realdp *P_parallel = D + k;
		//realdp *g_parallel = P_parallel + n * k;
		//realdp *Psig = g_parallel + k;
		//realdp *Lambda = new realdp[k + 1];
		//realdp *a_j = new realdp[k + 1];
		SolversTR::Run();
	};

	void LRTRSR1::tCG_TR(void)
	{/*The SolversTR:tCG_TR(void) is overridden. 
	 Though this method only makes sense when the Euclidean metric or an intrinsic approach is used,
	 it is much more efficient than the one in SolversTR:tCG_TR(void).*/

		integer n = gf1->Getlength(); /*number of rows*/
		integer k = Currentlength;    /*number of columns*/
		integer info = 0;
		
		integer idx;
		/*Compute Psi = Y - gamma S, PMGS = SY - gamma SS */
		for (integer i = 0; i < Currentlength; i++)
		{
			idx = (i + beginidx) % LengthSY;
			Mani->scalarVectorAddVector(x1, -gamma, S[idx], Y[idx], YMGS[i]);
			copy_(&n, const_cast<realdp *> (YMGS[i]->ObtainReadData()), &GLOBAL::IONE, Psi + i * n, &GLOBAL::IONE);
		}
		for (integer i = 0; i < Currentlength; i++)
		{
			for (integer j = 0; j < Currentlength; j++)
			{
				PMGQ[i + j * Currentlength] = SY[i + j * LengthSY] - gamma * SS[i + j * LengthSY];
			}
		}
		/*
		QR decomposion:           Psi = Q RT^T
		eigenvalue decomposition: U D U^T = eig(RT^T * PMGQ^{-1} RT)
		P_parallel = Q * U
		Therefore: Psi * (PMGQ)^{-1} * Psi = Q * U * D * U^T * Q^T = P_parallel * D * P_parallel^T
		*/
		if (k > 0)
		{
			gemm_(GLOBAL::T, GLOBAL::N, &k, &k, &n, &GLOBAL::DONE, Psi, &n, Psi, &n, &GLOBAL::DZERO, RT, &k); /*This can be avoid by using the previous computed Result.*/
			potrf_(GLOBAL::L, &k, RT, &k, &info);
			if (info != 0)
			{
				printf("Warning: cholesky decomposition LRTRSR1::tCG_TR fails with info:%d!\n", info);
			}
			for(integer i = 1; i < k; i++)
				for(integer j = 0; j < i; j++)
					RT[j + i * k] = 0;

			realdp *tmp = new realdp[k * k];
			integer k2 = k * k;
			copy_(&k2, PMGQ, &GLOBAL::IONE, LU_PMGQ, &GLOBAL::IONE);
			copy_(&k2, RT, &GLOBAL::IONE, tmp, &GLOBAL::IONE);

			/* Compute U D U^T = eig(RT^T * PMGQ^{-1} RT) */
			// LU decomposion for PMGQ, PMGQ = P * L * U, L and U are stored in PMGQ, the permutation matrix is in P
			// details: http://www.netlib.org/lapack/explore-html/d3/d6a/getrf_8f.html
			getrf_(&k, &k, LU_PMGQ, &k, P, &info);
			if (info < 0)
				printf("Error in forming Q matrix!\n");

			// solve linear system: PMGQ * X = v using the LU decomposition results from getrf, then solution is stored in v.
			// details: http://www.netlib.org/lapack/explore-html/d6/d49/getrs_8f.html
			getrs_(GLOBAL::N, &k, &k, LU_PMGQ, &k, P, tmp, &k, &info);

			gemm_(GLOBAL::T, GLOBAL::N, &k, &k, &k, &GLOBAL::DONE, RT, &k, tmp, &k, &GLOBAL::DZERO, U, &k);
			delete[] tmp;
			for (integer i = 0; i < k; i++)
			{
				for (integer j = i + 1; j < k; j++)
				{
					U[j + i * k] = (U[j + i * k] + U[i + j * k]) / 2;
					U[i + j * k] = U[j + i * k];
				}
			}
			integer lwork = -1;
			realdp lworkopt = 0;
			// eigvalue decomposition
			// find  the size of the memory required.
			syev_(GLOBAL::V, GLOBAL::U, &k, U, &k, D, &lworkopt, &lwork, &info);
			lwork = static_cast<integer> (lworkopt);
			realdp *work = new realdp[lwork];

			// compute eigvalue decomposition
			syev_(GLOBAL::V, GLOBAL::U, &k, U, &k, D, work, &lwork, &info);
			delete[] work;

			/*P_parallel = Psi * (RT^{-T} * U) */
			trtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &k, &k, RT, &k, U, &k, &info);
			if (info != 0)
			{
				printf("Warning: Solving a linear system LRTRSR1::tCG_TR fails with info:%d!\n", info);
			}
			gemm_(GLOBAL::N, GLOBAL::N, &n, &k, &k, &GLOBAL::DONE, Psi, &n, U, &k, &GLOBAL::DZERO, P_parallel, &n);
		}

		if (k > 0)
		{
			/*Psig = R^T Q^T g, the g_parallel is used as a temporary data here*/
			gemm_(GLOBAL::T, GLOBAL::N, &k, &GLOBAL::IONE, &n, &GLOBAL::DONE, Psi, &n, const_cast<realdp *> (gf1->ObtainReadData()), &n, &GLOBAL::DZERO, Psig, &k);
			/*g_parallel = P_parallel^T gf */
			gemm_(GLOBAL::T, GLOBAL::N, &k, &GLOBAL::IONE, &n, &GLOBAL::DONE, P_parallel, &n, const_cast<realdp *> (gf1->ObtainReadData()), &n, &GLOBAL::DZERO, g_parallel, &k);

		}

		//ForDebug::Print("Psi:", Psi, n, k);//----
		//ForDebug::Print("PMGQ", PMGQ, k, k);//---
		//std::cout << "gamma:" << gamma << std::endl;//---
		//ForDebug::Print("gf1:", gf1->ObtainReadData(), gf1->Getlength());
		//std::cout << "delta:" << Delta << std::endl;//---

		for (integer i = 0; i < k; i++)
			Lambda[i] = D[i] + gamma;
		Lambda[k] = gamma;
		for (integer i = 0; i < k + 1; i++)
			if (fabs(Lambda[i]) <= tolLocalNewton)
				Lambda[i] = 0;
		realdp lambda_min = (Lambda[0] < gamma) ? Lambda[0] : gamma;
		realdp a_kp2 = std::sqrt(dot_(&n, const_cast<realdp *> (gf1->ObtainReadData()), &GLOBAL::IONE, const_cast<realdp *> (gf1->ObtainReadData()), &GLOBAL::IONE) - dot_(&k, g_parallel, &GLOBAL::IONE, g_parallel, &GLOBAL::IONE));
		if (a_kp2 * a_kp2 < tolLocalNewton)
			a_kp2 = 0;

		for (integer i = 0; i < k; i++)
			a_j[i] = g_parallel[i];
		a_j[k] = a_kp2;

		realdp tmpv = 0;
		for (integer i = 0; i < k + 1; i++)
			tmpv += a_j[i] * a_j[i] / Lambda[i] / Lambda[i];
		tmpv = sqrt(tmpv);

		realdp *pStar = eta2->ObtainWriteEntireData();

		realdp sigmaStar = 0;
		tCGstatus = L_TR_MIN;
		if (lambda_min > 0 && tmpv <= Delta)
		{
			ComputeSBySMW(gamma, Psig, RT); /*gf1, Psig, YMGS, PMGQ, RT*/
		}
		else
			if (lambda_min <= 0 && PhiBar_fg(-lambda_min, Delta, Lambda, a_j) > 0)
			{
				sigmaStar = -lambda_min;
				realdp *v = new realdp[k + 1];
				for (integer i = 0; i < k + 1; i++)
				{
					if (fabs(Lambda[i] + sigmaStar) > tolLocalNewton)
					{
						v[i] = a_j[i] / (Lambda[i] + sigmaStar);
					}
					else
					{
						v[i] = 0;
					}
				}
				gemm_(GLOBAL::N, GLOBAL::N, &n, &GLOBAL::IONE, &k, &GLOBAL::DNONE, P_parallel, &n, v, &k, &GLOBAL::DZERO, pStar, &n);
				delete[] v;
				if (fabs(gamma + sigmaStar) > tolLocalNewton)
				{
					realdp *tmp = new realdp[n];
					gemm_(GLOBAL::N, GLOBAL::N, &n, &GLOBAL::IONE, &k, &GLOBAL::DONE, P_parallel, &n, g_parallel, &k, &GLOBAL::DZERO, tmp, &n);
					realdp coef = static_cast<realdp> (1) / (gamma + sigmaStar);
					axpy_(&n, &coef, tmp, &GLOBAL::IONE, pStar, &GLOBAL::IONE);
					coef = -coef;
					axpy_(&n, &coef, const_cast<realdp *> (gf1->ObtainReadData()), &GLOBAL::IONE, pStar, &GLOBAL::IONE);
					delete[] tmp;
				}

				if (lambda_min < 0)
				{
					realdp alpha = sqrt(Delta * Delta - dot_(&n, pStar, &GLOBAL::IONE, pStar, &GLOBAL::IONE));

					realdp *zstar = new realdp[n];
					for (integer i = 0; i < n; i++)
						zstar[i] = 0;
					if (fabs(lambda_min - Lambda[0]) < tolLocalNewton)
					{
						realdp coef = alpha / sqrt(dot_(&n, P_parallel, &GLOBAL::IONE, P_parallel, &GLOBAL::IONE));
						axpy_(&n, &coef, P_parallel, &GLOBAL::IONE, zstar, &GLOBAL::IONE);
					}
					else
					{
						realdp norm_umin = 1;
						for (integer i = 0; i < k; i++)
						{
							gemm_(GLOBAL::N, GLOBAL::T, &n, &GLOBAL::IONE, &k, &GLOBAL::DNONE, P_parallel, &n, P_parallel + i, &n, &GLOBAL::DZERO, zstar, &n);
							zstar[i] += 1;
							norm_umin = sqrt(dot_(&n, zstar, &GLOBAL::IONE, zstar, &GLOBAL::IONE));
							if (norm_umin > tolLocalNewton)
								break;
						}
						realdp coef = alpha / norm_umin;
						scal_(&n, &coef, zstar, &GLOBAL::IONE);
					}
					axpy_(&n, &GLOBAL::DONE, zstar, &GLOBAL::IONE, pStar, &GLOBAL::IONE);
					delete[] zstar;
				}
				tCGstatus = TR_NEGCURVTURE;
			}
			else
			{
				if (lambda_min > 0)
				{
					sigmaStar = LocalNewton(0, 100, tolLocalNewton, Lambda, a_j);
				}
				else
				{
					realdp sigmaHat = 0;
					for (integer i = 0; i < k + 1; i++)
					{
						if (sigmaHat < a_j[i] / Delta - Lambda[i])
							sigmaHat = a_j[i] / Delta - Lambda[i];
					}
					if (sigmaHat > -lambda_min)
						sigmaStar = LocalNewton(sigmaHat, 100, tolLocalNewton, Lambda, a_j);
					else
						sigmaStar = LocalNewton(-lambda_min, 100, tolLocalNewton, Lambda, a_j);
				}
				ComputeSBySMW(gamma + sigmaStar, Psig, RT); /*gf1, Psig, YMGS, PMGQ, RT*/
				tCGstatus = TR_EXCREGION;
			}
	};

	void LRTRSR1::ComputeSBySMW(realdp tauStar, realdp *Psig, realdp * RT)
	{/*gamma, gf1, Psig, YMGS, PMGQ, RT*/
		integer n = gf1->Getlength(); /*number of rows*/
		integer k = Currentlength;    /*number of columns*/
		realdp *pStar = eta2->ObtainWriteEntireData();
		realdp coef = static_cast<realdp> (-1) / tauStar;

		if (k != 0)
		{
			realdp *vw = new realdp[k * k];
			integer *Ptmp = new integer[k];
			integer k2 = k * k;
			gemm_(GLOBAL::N, GLOBAL::T, &k, &k, &k, &GLOBAL::DONE, RT, &k, RT, &k, &GLOBAL::DZERO, vw, &k);
			//ForDebug::Print("PsiPsi:", vw, k, k);//---
			scal_(&k2, &tauStar, vw, &GLOBAL::IONE);
			realdp tauStar2 = tauStar * tauStar;
			axpy_(&k2, &tauStar2, PMGQ, &GLOBAL::IONE, vw, &GLOBAL::IONE);
			//ForDebug::Print("vw:", vw, k, k);//---
			realdp *tmp = new realdp[k];
			integer info = 0;
			// LU decomposion for vw, vw = P * L * U, L and U are stored in vw, the permutation matrix is in Ptmp
			// details: http://www.netlib.org/lapack/explore-html/d3/d6a/getrf_8f.html
			getrf_(&k, &k, vw, &k, Ptmp, &info);
			if (info < 0)
				printf("Error in getrf of LRTRSR1::ComputeSBySMW\n");

			copy_(&k, Psig, &GLOBAL::IONE, tmp, &GLOBAL::IONE);
			// solve linear system: vw * X = tmp using the LU decomposition results from getrf, then solution is stored in tmp.
			// details: http://www.netlib.org/lapack/explore-html/d6/d49/getrs_8f.html
			getrs_(GLOBAL::N, &k, &GLOBAL::IONE, vw, &k, Ptmp, tmp, &k, &info);
			if (info < 0)
				printf("Error in getrs of LRTRSR1::ComputeSBySMW\n");

			gemm_(GLOBAL::N, GLOBAL::N, &n, &GLOBAL::IONE, &k, &GLOBAL::DONE, Psi, &n, tmp, &k, &GLOBAL::DZERO, pStar, &n);
			axpy_(&n, &coef, const_cast<realdp *> (gf1->ObtainReadData()), &GLOBAL::IONE, pStar, &GLOBAL::IONE);

			delete[] vw;
			delete[] tmp;
			delete[] Ptmp;
			return;
		}
		gf1->CopyTo(eta2);
		scal_(&n, &coef, pStar, &GLOBAL::IONE);
	};

	realdp LRTRSR1::PhiBar_fg(realdp nlambda_min, realdp Delta, realdp *Lambda, realdp *a_j, realdp *gf)
	{
		//integer n = gf1->Getlength(); /*number of rows*/
		integer k = Currentlength;    /*number of columns*/
		realdp *tmp = new realdp[k + 1];
		for (integer i = 0; i < k + 1; i++)
			tmp[i] = Lambda[i] + nlambda_min;
		realdp eps_tol = static_cast<realdp> (1e-10);
		realdp gradf = 0;

		bool flag = false;
		for (integer i = 0; i < k + 1; i++)
		{
			if (fabs(a_j[i]) < eps_tol || fabs(tmp[i]) < eps_tol)
			{
				flag = true;
				break;
			}
		}
		if (flag)
		{
			realdp pnorm2 = 0;
			for (integer i = 0; i < k + 1; i++)
			{
				if (fabs(a_j[i]) > eps_tol && fabs(tmp[i]) < eps_tol)
				{
					gradf = static_cast<realdp> (1) / eps_tol;
					if (gf != nullptr)
						*gf = gradf;
					delete[] tmp;
					return static_cast<realdp> (-1) / Delta;
				}
				else
				if (fabs(a_j[i]) > eps_tol && fabs(tmp[i]) > eps_tol)
				{
					pnorm2 += a_j[i] * a_j[i] / tmp[i] / tmp[i];
					gradf += a_j[i] * a_j[i] / tmp[i] / tmp[i] / tmp[i];
				}
			}
			realdp norm = sqrt(pnorm2);
			gradf = gradf / norm / norm / norm;
			if (gf != nullptr)
				*gf = gradf;
			delete[] tmp;
			return static_cast<realdp> (1) / norm - static_cast<realdp> (1) / Delta;
		}

		for (integer i = 0; i < k + 1; i++)
			gradf += a_j[i] * a_j[i] / tmp[i] / tmp[i] / tmp[i];

		for (integer i = 0; i < k + 1; i++)
			tmp[i] = a_j[i] / tmp[i];
		integer length = k + 1;
		realdp norm = sqrt(dot_(&length, tmp, &GLOBAL::IONE, tmp, &GLOBAL::IONE));

		gradf = gradf / norm / norm / norm;
		if (gf != nullptr)
			*gf = gradf;
		delete[] tmp;
		return static_cast<realdp> (1) / norm - static_cast<realdp> (1) / Delta;
	};

	realdp LRTRSR1::LocalNewton(realdp x0, integer maxIter, realdp tol, realdp *Lambda, realdp *a_j)
	{
		integer k = 0;
		realdp gf = 0;
		realdp fv = PhiBar_fg(x0, Delta, Lambda, a_j, &gf);
		while (fabs(fv) > std::numeric_limits<realdp>::epsilon() && k < maxIter)
		{
			x0 -= fv / gf;
			fv = PhiBar_fg(x0, Delta, Lambda, a_j, &gf);
			k++;
		}
		return x0;
	};

	void LRTRSR1::CheckParams(void)
	{
		SolversTR::CheckParams();
		char YES[] = "YES";
		char NO[] = "NO";
		char *status;

		printf("LRTRSR1 METHOD PARAMETERS:\n");
		status = YES;
		printf("isconvex      :%15d[%s],\t", isconvex, status);
		status = (LengthSY >= 0) ? YES : NO;
		printf("LengthSY      :%15d[%s]\n", LengthSY, status);
	};

	void LRTRSR1::HessianEta(Vector *Eta, Vector *result)
	{
		HvLRTRSR1(Eta, YMGS, LU_PMGQ, P, Currentlength, gamma, result);
	};

	void LRTRSR1::UpdateData(void)
	{
		UpdateDataLRTRSR1(s, y, Py, S, Y, SS, SY, PMGQ, LU_PMGQ, P, Currentlength, inpss, inpsy, inpyy, gamma, beginidx, isupdated);
	};

	void LRTRSR1::Acceptence(void)
	{
		for (integer i = 0; i < Currentlength; i++)
		{
			Mani->VectorTransport(x1, eta2, x2, S[i], S[i]);
			Mani->VectorTransport(x1, eta2, x2, Y[i], Y[i]);
		}
	};

	void LRTRSR1::PrintGenInfo(void)
	{
		Solvers::PrintGenInfo();
		printf("nH:%d,rho:%.2e,radius:%.3e,tCGstatus:%s,", nH, rho, Delta, tCGstatusSetnames[tCGstatus].c_str());
	};

	void LRTRSR1::PrintInfo(void)
	{
		printf("\n\tgamma:%.3e,inpss:%.3e,inpsy:%.3e,inpyy:%.3e,IsUpdateHessian:%d,", gamma, inpss, inpsy, inpyy, isupdated);
		printf("\n");
	};
}; /*end of ROPTLIB namespace*/
