
#include "Solvers/RPG.h"

/*Define the namespace*/
namespace ROPTLIB{

	RPG::RPG(const Problem *prob, const Variable *initialx, const Variable *insoln)
	{
		Initialization(prob, initialx, insoln);
	};

    RPG::~RPG()
    {
        delete[] Fu;
        delete VecTMP1;
        delete VecTMP2;
        delete BLambda;
        delete DLambda;
        if(Weight != nullptr)
            delete Weight;
    };

	void RPG::SetProbX(const Problem *prob, const Variable *initialx, const Variable *insoln)
	{
		SolversLS::SetProbX(prob, initialx, insoln);
		prob->SetUseGrad(true);
		prob->SetUseHess(false);
        VecTMP1 = prob->GetDomain()->GetEMPTYEXTR()->ConstructEmpty();
        VecTMP2 = prob->GetDomain()->GetEMPTYEXTR()->ConstructEmpty();
        Weight = nullptr;
        LengthW = initialx->Getlength();
        BLambda = prob->GetDomain()->GetEMPTYEXTR()->ConstructEmpty();
        DLambda = prob->GetDomain()->GetEMPTYEXTR()->ConstructEmpty();
        dimNorVec = prob->GetDomain()->GetExtrDim() - prob->GetDomain()->GetIntrDim();
        
        Fu = new realdp[dimNorVec * 6];
        Fz = Fu + dimNorVec;
        initDx = Fz + dimNorVec;
        SMz = initDx + dimNorVec;
        SMd = SMz + dimNorVec;
        SMu = SMd + dimNorVec;
        for (integer i = 0; i < dimNorVec; i++)
        {
            Fu[i] = 0;
            Fz[i] = 0;
            initDx[i] = 0;
            SMz[i] = 0;
            SMd[i] = 0;
            SMu[i] = 0;
        }
	};

	void RPG::SetDefaultParams()
	{
		SolversLS::SetDefaultParams();
		InitSteptype = ONESTEP;
        Minstepsize = 2e-2;
        LS_ratio1 = 0.5;
        LS_ratio2 = 0.5;
        SMlambda = 0.2;
        SMtol = 1e-10;
        Tolerance = 1e-4;
        adavalue = 1;
        Variant = ADALIPSCHITZ;
        totalSMiter = 0;
        totalSMCGiter = 0;
		SolverName.assign("RPG");
	};

    void RPG::PrintInfo(void)
    {
        printf("SMlambda:%.3e,SMtol:%.3e,SMiter:%d,SMCGiter:%d", SMlambda, SMtol, SMiter, SMCGiter);
        printf("\n");
    };

    void RPG::PrintGenInfo(void)
    {
        printf("i:%d,f:%.16e,df/f:%.3e,", iter, f2, ((f1 - f2) / std::fabs(f2)));

        printf("nsubgf:%.3e,", nsubgf);

        printf("time:%.2e,", static_cast<realdp>(getTickCount() - starttime) / CLK_PS);

        printf("nf:%d,ng:%d,nR:%d,", nf, ng, nR);

        if (nV != 0)
            printf("nV(nVp):%d(%d),", nV, nVp);
        
        printf("LSstatus:%s,", LSstatusSetnames[LSstatus].c_str());
        if (newslope != 0)
        {
            printf("newslope:%.3e,", newslope);
        }
        printf("initstepsize:%.3e,stepsize:%.3e,", initiallength, stepsize);
    };

    void RPG::CheckParams(void)
    {
        SolversLS::CheckParams();

        std::string RPGVariantnames[RPGVARIANTLENGTH] = { "REGULAR", "ADALIPSCHITZ" };

        char YES[] = "YES";
        char NO[] = "NO";
        char *status;

        printf("RPG-based METHODS PARAMETERS:\n");
        status = (LengthW > 0 ) ? YES : NO;
        printf("LengthW       :%15d[%s]\t", LengthW, status);
        status = (Variant >= 0 && Variant < RPGVARIANTLENGTH) ? YES : NO;
        printf("RPGVariant    :%15s[%s]\n", RPGVariantnames[Variant].c_str(), status);
    };

    void RPG::Run(void)
    {
        if(Weight != nullptr)
        {
            delete Weight;
            Weight = nullptr;
        }
        Weight = new EucVector(LengthW);
        
        bool nonmonoticLS = false;
        Variable *xTemp;
        Vector *gfTemp;
        pre_funs.clear();

        Solvers::Run();

        ChooseLinesearch();

        LSstatus = SUCCESS;
        f1 = Prob->f(x1); nf++;
        f2 = f1;
        Prob->Grad(x1, gf1); ng++;
        ngf0 = sqrt(Mani->Metric(x1, gf1, gf1));
        ngf1 = ngf0; ngf2 = ngf1;
        newslope = 0;
        iter = 0;
        realdp *initialstepsizeptr = nullptr, *acceptedstepsizeptr = nullptr;
        if (Debug >= FINALRESULT)
            printf("i:%d,f:%.3e,|gf|:%.3e,\n", iter, f1, ngf1);
            
        if (Debug >= ITERRESULT)
        {
            timeSeries[iter] = static_cast<realdp>(getTickCount() - starttime) / CLK_PS;
            funSeries[iter] = f1;
            gradSeries[iter] = ngf1;
            distSeries[iter] = ((soln == nullptr) ? 0 : Mani->Dist(x1, soln));
            initialstepsizeptr = new realdp[1 + Max_Iteration];
            acceptedstepsizeptr = new realdp[1 + Max_Iteration];
            initialstepsizeptr[iter] = 0;
            acceptedstepsizeptr[iter] = 0;
        }
        
        bool isstop = IsStopped();
        /*If the intrinsic representation is used, then exeta1 and exeta2 are used to store the extrinsic representations of eta1 and eta2 respectively*/
        if (Prob->GetDomain()->GetIsIntrinsic())
        {
            exeta1 = Prob->GetDomain()->GetEMPTYEXTR()->ConstructEmpty();
            exeta2 = Prob->GetDomain()->GetEMPTYEXTR()->ConstructEmpty();
        }

        /*Start the loop*/
        while ((((! isstop) && iter < Max_Iteration) || iter < Min_Iteration) && LSstatus == SUCCESS)
        {
            GetSearchDir(); // Obtain search direction eta1
            initialslope = Mani->Metric(x1, gf1, eta1);
            /*Compute initial step size for the next iteration*/
            InitialStepSize();
            initiallength = stepsize;
            if (Debug >= ITERRESULT)
                initialstepsizeptr[iter + 1] = initiallength;
            /*Start a line search algorithm. If the intrinsic representation is used for the search direction, then converting it into the extrinsic representation.*/
            if (Prob->GetDomain()->GetIsIntrinsic())
            {
                Prob->GetDomain()->ObtainExtr(x1, eta1, exeta1);
            }
            
            /*If accurate enough, then a fixed stepsize is chosen.*/
            if (ngf1 / (ngf0 + Tolerance) < Accuracy)
            {
                if (Finalstepsize > 0)
                    stepsize = Finalstepsize;
                else
                    stepsize = initiallength;
                f2 = h();
                Prob->Grad(x2, gf2); ng++;
            }
            else
            {
                /* Call the specified linesearch algorithm.
                Note that in the linesearch algorithm, we need to obtain
                accepted stepsize, eta2=stepsize*eta1, x2 = R_{x_1}(eta_2), f2 = f(x2), and gf2 = grad f(x_2) */
                if (LineSearch_LS == INPUTFUN)
                {
                    if (Prob->GetDomain()->GetIsIntrinsic())
                    {
                        stepsize = LinesearchInput(iter, x1, exeta1, stepsize, initialslope, Prob, this);
                    }
                    else
                    {
                        stepsize = LinesearchInput(iter, x1, eta1, stepsize, initialslope, Prob, this);
                    }
                    stepsize = (stepsize < std::numeric_limits<realdp>::epsilon()) ? initiallength : stepsize;
                    initiallength = stepsize;
                    if (!IsPureLSInput)
                    {
                        SolversLS::LinesearchArmijo();
                    }
                    else
                    {
                        f2 = h();
                        Prob->Grad(x2, gf2); ng++;
                    }
                }
                else
                {
                    (this->*Linesearch)();
                    if(LSstatus == MINSTEPSIZE)
                    {
                        if(SMlambda == 1e-10 && SMtol == 1e-24)
                        {
                            break;
                        }
                        SMtol = ((SMtol * 0.01 > 1e-24) ? SMtol * 0.01 : 1e-24); //--std::max(SMtol * 0.01, 1e-24);
                        SMlambda = ((SMlambda * 0.1 > 1e-10) ? SMlambda * 0.1 : 1e-10); //--std::max(SMlambda * 0.1, 1e-10);
                        LSstatus = SUCCESS;
                        continue;
                    }
                }
            }
            if(Variant == ADALIPSCHITZ)
            {
                if(stepsize == initiallength)
                {
                    adavalue /= 1.01;
                }
                else
                {
                    adavalue = ((adavalue * 1.01 < 1.0) ? adavalue * 1.01 : 1.0);//-- std::min(adavalue * 1.01, 1.0);
                }
            }
            
            /*Output debug information if necessary.*/
            if (LSstatus < SUCCESS && Debug > FINALRESULT)
            {
                printf("Linesearch fails! LSstatus:%s\n", LSstatusSetnames[LSstatus].c_str());
            }

            if (std::isnan(f2) || std::isinf(f2)) /*Stop when got a nan or inf*/
            {
                printf("New function value is either nan or inf. Stop!\n");
                break;
            }

            iter++;

            /*Update the Hessian approximation for quasi-Newton methods
                or obtain search direction candadite for Riemannian nonlinear conjugate gradient*/
            UpdateData();

            /*norm of the gradient at x2*/
            ngf2 = sqrt(Mani->Metric(x2, gf2, gf2));
            
            /*Call the function to check whether the stopping criterion is satisfied or not.
            The default function is written in Solvers.h and Solvers.cpp*/
            isstop = IsStopped();

            if (Debug >= ITERRESULT)
            {
                /*Output information*/
                if (iter % OutputGap == 0)
                {
                    PrintGenInfo();
                    PrintInfo(); // Output information specific to Algorithms
                }
                /*Store debug information in the arrays*/
                timeSeries[iter] = static_cast<realdp>(getTickCount() - starttime) / CLK_PS;
                funSeries[iter] = f2; gradSeries[iter] = ngf2;
                distSeries[iter] = ((soln == nullptr) ? 0 : Mani->Dist(x2, soln));
                acceptedstepsizeptr[iter] = stepsize;
            }

            /*Switch information at x1 and x2*/
            xTemp = x1; x1 = x2; x2 = xTemp;
            gfTemp = gf1; gf1 = gf2; gf2 = gfTemp;
            if(ngf2 / (ngf0 + Tolerance) <= PreFunsAccuracy)
                nonmonoticLS = true;
            
            pre_funs.push_front(f1);
            if ((pre_funs.size() > Num_pre_funs || !nonmonoticLS) && pre_funs.size() > 1)
                pre_funs.pop_back();
            f1 = f2; // This need be after the update of pre_funs. It is used in initialstepsize().
            ngf1 = ngf2;
        }
        if (Prob->GetDomain()->GetIsIntrinsic())
        {
            delete exeta1;
            delete exeta2;
        }
        ComTime = static_cast<realdp>(getTickCount() - starttime) / CLK_PS;
        if (Debug >= ITERRESULT)
        {
            lengthSeries = iter + 1;
            SharedSpace *sharedinitialstepsize = new SharedSpace(1, lengthSeries);
            SharedSpace *sharedacceptedstepsize = new SharedSpace(1, lengthSeries);
            realdp *tmp1 = sharedinitialstepsize->ObtainWriteEntireData();
            realdp *tmp2 = sharedacceptedstepsize->ObtainWriteEntireData();
            copy_(&lengthSeries, initialstepsizeptr, &GLOBAL::IONE, tmp1, &GLOBAL::IONE);
            copy_(&lengthSeries, acceptedstepsizeptr, &GLOBAL::IONE, tmp2, &GLOBAL::IONE);

            SolverTMP->AddToTempData("initialstepsize", sharedinitialstepsize);
            SolverTMP->AddToTempData("acceptedstepsize", sharedacceptedstepsize);
            delete[] initialstepsizeptr;
            delete[] acceptedstepsizeptr;
        }
        PrintFinalInfo();
        if (nsubgf != -1)
        {
            ngf1 = nsubgf;
        }
    };

    void RPG::PrintFinalInfo(void)
    {
        if (Debug >= FINALRESULT)
        {
            printf("Iter:%d,f:%.16e,", iter, f1);
            printf("nsubgf:%.3e,", nsubgf);
            
            printf("time:%.2e,", ComTime);
            if (subprobtimes != 0)
            {
                printf("nsubprob:%d,", subprobtimes);
            }
            printf("nf:%d,ng:%d,nR:%d,", nf, ng, nR);
            
            if (nH != 0)
            {
                printf("nH:%d,", nH);
            }
            if (nV != 0)
            {
                printf("nV(nVp):%d(%d),", nV, nVp);
            }
            printf("SMlambda:%.3e,SMtol:%.3e,totalSMiter:%d,totalSMCGiter:%d", SMlambda, SMtol, totalSMiter, totalSMCGiter);
            Solvers::PrintFinalInfo();
        }
    };

	void RPG::GetSearchDir(void)
	{
        ProximalMap(x1, gf1, initDx, eta1);
	};

    void RPG::ProximalMap(Variable *x, Vector *gfx, realdp *initD, Vector *eta)
    { /*eta = argmin g(gfx, eta) + 0.5 \|eta\|_W^2 + h(eta), W is either a scalar or a weight matrix that is given by function "PreConditioner" defined in Problem.
       Only extrinsic representation is supported. */
        /*parameters*/
        realdp SMtau = 0.1, SMalpha = 0.1;
        integer SMmaxiter = 100, maxSMbtiter = 10;
        std::string status;
        integer SMbtiter = 0;
        bool earlytermination = false;
        Prob->PreConditioner(x, nullptr, Weight);
        realdp *Wptr = Weight->ObtainWritePartialData();
        
        if(Variant == ADALIPSCHITZ)
            scal_(&LengthW, &adavalue, Wptr, &GLOBAL::IONE);
        
        realdp nFz = 0, nFu = 0, mu = 0;
        copy_(&dimNorVec, initD, &GLOBAL::IONE, SMz, &GLOBAL::IONE);
//        printf("dimNorVec:%d\n", dimNorVec);//--
//        ForDebug::Print("SMz:", SMz, dimNorVec);//---
        ComputeBLambda(x, SMz, gfx);
//        BLambda->Print("BLambda init:");//---
        EW(x, Fz);
//        ForDebug::Print("Fz:", Fz, dimNorVec);//----
        nFz = nrm2_(&dimNorVec, Fz, &GLOBAL::IONE);
//        fvs[0] = nFz;
        for(integer i = 0; i < dimNorVec; i++)
            SMd[i] = 0;
        SMiter = 0;
//        std::cout << "nFz initial:" << nFz << ", SMiter:" << SMiter << std::endl;//---
        while(nFz * nFz > SMtol && SMiter < SMmaxiter)
        {
            mu = ((nFz < 0.1) ? nFz : 0.1);
            mu = ((mu > 1e-11) ? mu : 1e-11);
            mu = SMlambda *  mu; //--std::max(std::min(nFz, 0.1), 1e-11);
            SMCGiter = myCG(x, Fz, mu, SMtau, SMlambda * nFz, ((dimNorVec < 30) ? dimNorVec : 30) , SMd); //--std::min(dimNorVec, 30)
//            ForDebug::Print("SMd:", SMd, dimNorVec);//----
            copy_(&dimNorVec, SMd, &GLOBAL::IONE, SMu, &GLOBAL::IONE);
            axpy_(&dimNorVec, &GLOBAL::DONE, SMz, &GLOBAL::IONE, SMu, &GLOBAL::IONE);
            ComputeBLambda(x, SMu, gfx);
            EW(x, Fu);
            nFu = nrm2_(&dimNorVec, Fu, &GLOBAL::IONE);
            
            SMalpha = 1;
            SMbtiter = 0;
            while(nFu * nFu > nFz * nFz * (1 - 0.001 * SMalpha) && SMbtiter < maxSMbtiter)
            {
                SMalpha *= 0.5;
                copy_(&dimNorVec, SMz, &GLOBAL::IONE, SMu, &GLOBAL::IONE);
                axpy_(&dimNorVec, &SMalpha, SMd, &GLOBAL::IONE, SMu, &GLOBAL::IONE);
                ComputeBLambda(x, SMu, gfx);
                EW(x, Fu);
                nFu = nrm2_(&dimNorVec, Fu, &GLOBAL::IONE);
                SMbtiter++;
            }
//            std::cout << "nFu:" << nFu << ",nFz:" << nFz << std::endl;//---
//            std::cout << "nFu * nFu:" << nFu * nFu << ",nFz * nFz * (1 - 0.001 * SMalpha):" << nFz * nFz * (1 - 0.001 * SMalpha) << std::endl;//---
//            std::cout << "nFu * nFu:" << nFu * nFu << ",nFz * nFz * (1 - 0.001 * SMalpha):" << nFz * nFz * (1 - 0.001 * SMalpha) << std::endl;//---
            /*We let the SSN run for at least 1 iteration before early termination*/
            if(SMbtiter == maxSMbtiter && SMiter != 0)
            {
                earlytermination = true;
//                std::cout << "early terminination" << std::endl;//---
                break;
            }
            copy_(&dimNorVec, SMu, &GLOBAL::IONE, SMz, &GLOBAL::IONE);
            copy_(&dimNorVec, Fu, &GLOBAL::IONE, Fz, &GLOBAL::IONE);
            nFz = nFu;
            SMiter++;
//            std::cout << "nFz:" << nFz << ", SMiter:" << SMiter << std::endl;//---
        }
//        std::cout << "nFz end:" << nFz << ", SMiter:" << SMiter << std::endl;//---
        copy_(&dimNorVec, SMz, &GLOBAL::IONE, initD, &GLOBAL::IONE);
//        BLambda->Print("BLambda end:");//---
//        Weight->Print("Weight:");//---
        Prob->ProxW(BLambda, Weight, eta);
        Prob->GetDomain()->VectorMinusVector(nullptr, eta, x, eta);
//        eta1->Print("eta1 end:");//---
        totalSMiter += SMiter;
        totalSMCGiter += SMCGiter;
    };

    realdp RPG::maxvec(realdp *vec, integer length)
    {
        realdp maxv = -100000000000;
        for(integer i = 0; i < length; i++)
        {
            if(vec[i] > maxv)
                maxv = vec[i];
        }
        return maxv;
    };

    integer RPG::myCG(Variable *xx, realdp *nb, realdp mu, realdp tau, realdp lambdanFz, integer maxiter, double *d)
    {
        realdp *x, *r, *p, *Ap;
        realdp alpha = 0, tmp = 0, beta = 0, rr0 = 0;
        x = new realdp[4 * dimNorVec];
        r = x + dimNorVec;
        p = r + dimNorVec;
        Ap = p + dimNorVec;
//        ForDebug::Print("x0:", SMd, dimNorVec);//---
        GLdW(xx, SMd, Ap);
//        ForDebug::Print("Ax0 1:", Ap, dimNorVec);//---
        axpy_(&dimNorVec, &mu, SMd, &GLOBAL::IONE, Ap, &GLOBAL::IONE);
//        ForDebug::Print("Ax0 2:", Ap, dimNorVec);//---
        for(integer i = 0; i < dimNorVec; i++)
        {
            x[i] = SMd[i];
            r[i] = -nb[i] - Ap[i];
            p[i] = r[i];
        }
        integer CGiter = 0;
        while(nrm2_(&dimNorVec, r, &GLOBAL::IONE) > tau * ((lambdanFz * nrm2_(&dimNorVec, x, &GLOBAL::IONE) < 1.0) ? lambdanFz * nrm2_(&dimNorVec, x, &GLOBAL::IONE) : 1.0) && CGiter < maxiter) //std::min(lambdanFz * nrm2_(&dimNorVec, x, &GLOBAL::IONE), 1.0)
        {
//            ForDebug::Print("p:", p, dimNorVec);//---
            GLdW(xx, p, Ap);
//            ForDebug::Print("Ap:", Ap, dimNorVec);//---
            axpy_(&dimNorVec, &mu, p, &GLOBAL::IONE, Ap, &GLOBAL::IONE);
//            ForDebug::Print("Ap:", Ap, dimNorVec);//---
            rr0 = dot_(&dimNorVec, r, &GLOBAL::IONE, r, &GLOBAL::IONE);
            alpha = rr0 / dot_(&dimNorVec, p, &GLOBAL::IONE, Ap, &GLOBAL::IONE);
            axpy_(&dimNorVec, &alpha, p, &GLOBAL::IONE, x, &GLOBAL::IONE);
            tmp = - alpha;
            axpy_(&dimNorVec, &tmp, Ap, &GLOBAL::IONE, r, &GLOBAL::IONE);
            beta = dot_(&dimNorVec, r, &GLOBAL::IONE, r, &GLOBAL::IONE) / rr0;
            scal_(&dimNorVec, &beta, p, &GLOBAL::IONE);
            axpy_(&dimNorVec, &GLOBAL::DONE, r, &GLOBAL::IONE, p, &GLOBAL::IONE);
            CGiter++;
        }
        dcopy_(&dimNorVec, x, &GLOBAL::IONE, d, &GLOBAL::IONE);
//        std::cout << "res:" << nrm2_(&dimNorVec, r, &GLOBAL::IONE) << std::endl;//--
        delete[] x;
        return CGiter;
    };

	void RPG::ComputeBLambda(Variable *x, realdp *d, Vector *gfx)
	{
        calAadj(x, d, VecTMP1);
//        VecTMP1->Print("A^* d:");//---
        Prob->GetDomain()->VectorMinusVector(nullptr, gfx, VecTMP1, VecTMP1);
//        VecTMP1->Print("gf1 - A^* d:");//---
//        Weight->Print("W:");//----
        integer nblock = Weight->Getlength();
        integer blocksize = x->Getlength() / nblock;
        realdp *Vec1ptr = VecTMP1->ObtainWritePartialData();
        
        for(integer i = 0; i < nblock; i++)
        {
            const realdp L = Weight->ObtainReadData()[i];
            realdp tmp = 1.0 / L;
            scal_(&blocksize, &tmp, Vec1ptr + i * blocksize, &GLOBAL::IONE);
        }
        
//        if(Weight->Getlength() == 1)
//        {
//            const realdp L = Weight->ObtainReadData()[0];
//            realdp tmp = 1.0 / L;
//            integer length = VecTMP1->Getlength();
//            realdp *Vec1ptr = VecTMP1->ObtainWritePartialData();
//            scal_(&length, &tmp, Vec1ptr, &GLOBAL::IONE);
//        }else
//        {
//            Prob->GetDomain()->VectorHadamardDivideVector(nullptr, VecTMP1, Weight, VecTMP1);
//        }
        Prob->GetDomain()->VectorMinusVector(nullptr, x, VecTMP1, BLambda);
	};

	void RPG::EW(Variable *x, realdp *ELambda)
	{
		Prob->ProxW(BLambda, Weight, DLambda);
//        DLambda->Print("DLambda 1:");//---
		Prob->GetDomain()->VectorMinusVector(nullptr, DLambda, x, DLambda);
//        DLambda->Print("DLambda 2:");//---
		calA(x, DLambda, ELambda);
	};

	void RPG::GLdW(Variable *x, realdp *d, realdp *GLambda)
	{
//        ForDebug::Print("d:", d, dimNorVec);//---
		calAadj(x, d, VecTMP1);
//        VecTMP1->Print("VecTMP1:");//---
		Prob->CalJW(BLambda, VecTMP1, Weight, VecTMP2);
//        VecTMP2->Print("VecTMP2:");//---
        
        
        integer nblock = Weight->Getlength();
        integer blocksize = x->Getlength() / nblock;
        realdp *Vec2ptr = VecTMP2->ObtainWritePartialData();
        
        for(integer i = 0; i < nblock; i++)
        {
            const realdp L = Weight->ObtainReadData()[i];
            realdp tmp = 1.0 / L;
            scal_(&blocksize, &tmp, Vec2ptr + i * blocksize, &GLOBAL::IONE);
        }
//        if(Weight->Getlength() == 1)
//        {
//            const realdp L = Weight->ObtainReadData()[0];
//            realdp tmp = 1.0 / L;
//            integer length = VecTMP2->Getlength();
//            realdp *Vec2ptr = VecTMP2->ObtainWritePartialData();
//            scal_(&length, &tmp, Vec2ptr, &GLOBAL::IONE);
//        }else
//        {
//            Prob->GetDomain()->VectorHadamardDivideVector(nullptr, VecTMP2, Weight, VecTMP2);
//        }
		calA(x, VecTMP2, GLambda);
	};

	void RPG::calA(Variable *x, Vector *DLambda, realdp *ELambda)
	{
		Prob->GetDomain()->ObtainNorVerIntr(x, DLambda, ELambda);
        /*for debug*/
        //testCalACalAadj(x, DLambda, ELambda);
	};

	void RPG::calAadj(Variable *x, realdp *d, Vector *CalAadjd)
	{
		Prob->GetDomain()->ObtainNorVerExtr(x, d, CalAadjd);
	};

    bool RPG::IsStopped(void)
    {
        const realdp *eta1ptr = eta1->ObtainReadData();
        
        
        integer nblock = Weight->Getlength();
        integer blocksize = eta1->Getlength() / nblock;

        const realdp *Wptr = Weight->ObtainReadData();
        if(eta1ptr != nullptr && (Wptr != nullptr))
        {
            nsubgf = 0;
            for(integer i = 0; i < nblock; i++)
            {
                for(integer j = 0; j < blocksize; j++)
                    nsubgf += eta1ptr[j + i * blocksize] * eta1ptr[j + i * blocksize] * Wptr[i] * Wptr[i];
            }
            nsubgf = std::sqrt(nsubgf);
        }
        
//        integer length = eta1->Getlength();
//        if(Weight->Getlength() != 1)
//        {
//            const realdp *Wptr = Weight->ObtainReadData();
//            if(eta1ptr != nullptr && (Wptr != nullptr))
//            {
//                nsubgf = 0;
//                for(integer i = 0; i < length; i++)
//                    nsubgf += eta1ptr[i] * eta1ptr[i] * Wptr[i] * Wptr[i];
//                nsubgf = std::sqrt(nsubgf);
//            }
//        } else
//        {
//            const realdp *Weightptr = Weight->ObtainReadData();
//            if(eta1ptr != nullptr && Weightptr != nullptr)
//            {
//                nsubgf = 0;
//                for(integer i = 0; i < length; i++)
//                    nsubgf += eta1ptr[i] * eta1ptr[i] * Weightptr[0] * Weightptr[0];
//                nsubgf = std::sqrt(nsubgf);
//            }
//        }
        
        return (nsubgf < Tolerance) && (nsubgf != -1);
    };

    void RPG::testCalACalAadj(Variable *x, Vector *DLambda, realdp *ELambda)
    {
        Prob->GetDomain()->ObtainNorVerIntr(x, DLambda, ELambda);
        realdp *tmpp = new realdp[dimNorVec];
        for(integer i = 0; i < dimNorVec; i++)
            tmpp[i] = genrandnormal();
        std::cout << "<A d, t>:" << dot_(&dimNorVec, ELambda, &GLOBAL::IONE, tmpp, &GLOBAL::IONE) << std::endl;
        
        Vector *Vtmp = DLambda->ConstructEmpty();
        calAadj(x, tmpp, Vtmp);
        std::cout << "<d, A^* t>:" << Prob->GetDomain()->Metric(nullptr, Vtmp, DLambda) << std::endl;
        delete Vtmp;
        delete[] tmpp;
    }

    void RPG::SetParams(PARAMSMAP params)
    {
        SolversLS::SetParams(params);
        PARAMSMAP::iterator iter;
        for (iter = params.begin(); iter != params.end(); iter++)
        {
            if (iter->first == static_cast<std::string> ("RPGVariant"))
            {
                Variant = static_cast<RPGVariant> (static_cast<integer> (iter->second));
            }
            if (iter->first == static_cast<std::string> ("LengthW"))
            {
                LengthW = static_cast<integer> (iter->second);
            }
        }
    };
}; /*end of ROPTLIB namespace*/
