/*
This file defines the proximal gradient method on manifold with or without preconditioner.


Solvers --> QuasiNewton --> SolversLS --> RPG

---- WH
*/

#ifndef RPG_H
#define RPG_H

#include "Solvers/SolversLS.h"
#include "Others/def.h"

/*Define the namespace*/
namespace ROPTLIB{

enum RPGVariant{ REGULAR, ADALIPSCHITZ, RPGVARIANTLENGTH };

	class RPG : public SolversLS{
	public:
		/*The contructor of RPG method. It calls the function Solvers::Initialization.
		INPUT : prob is the problem which defines the cost function, gradient and possible the action of Hessian
		and specifies the manifold of domain.
		initialx is the initial iterate.
		insoln is the true solution. It is not required and only used for research.*/
		RPG(const Problem *prob, const Variable *initialx, const Variable *insoln = nullptr);

        ~RPG();
        
		/*Call Solvers::SetProbX function and indicate RCG does not need action of Hessian.
		INPUT:	prob is the problem which defines the cost function, gradient and possible the action of Hessian
		and specifies the manifold of domain.
		initialx is the initial iterate.
		insoln is the true solution. It is not required and only used for research.*/
		virtual void SetProbX(const Problem *prob, const Variable *initialx, const Variable *insoln);

		/*Setting parameters (member variables) to be default values */
		virtual void SetDefaultParams();
        
        /*Check whether the parameters in RPG are legal or not.*/
        virtual void CheckParams(void);
        
        /*Print general information, which is not specific to an algorithm.*/
        virtual void PrintGenInfo(void);
        
        /*Print information specific to an algorithm*/
        virtual void PrintInfo(void);

        /*Run the proximal gradient-based algorithm.*/
        void Run(void);
        
        /*PARAMSMAP is defined in "def.h" and it is a map from string to realdp, i.e., std::map<std::string, realdp> .
        This function is used to set the parameters by the mapping*/
        virtual void SetParams(PARAMSMAP params);
        
        RPGVariant Variant;
        
        integer LengthW;
	protected:
        /*Print last information in an algorithm*/
        virtual void PrintFinalInfo(void);

        /*Check whether a stopping criterion is satisfied or not*/
        virtual bool IsStopped(void);
        
		/*Set the search direction computed by semi-smooth Newton method
		The direction is
		argmin_{d \in \T_x \mathcal{M}} <gfx, d> + (1/2) \|d\|_W^2 + lambda \|x + d\|_1,
		where \|d\|_W^2 = \vec(d)^T W \vec(d), and W is a preconditioner. */
		virtual void GetSearchDir();

        virtual void ProximalMap(Variable *x, Vector *gfx, realdp *initD, Vector *eta);
        
		virtual void ComputeBLambda(Variable *x, realdp *d, Vector *gfx);

		virtual void EW(Variable *x, realdp *ELambda);

		virtual void GLdW(Variable *x, realdp *d, realdp *GLambda);

		virtual void calA(Variable *x, Vector *DLambda, realdp *ELambda);

		virtual void calAadj(Variable *x, realdp *d, Vector *CalAadjd);
        
        integer myCG(Variable *x, realdp *nb, realdp mu, realdp tau, realdp lambdanFz, integer maxiter, realdp *d);
        
        realdp maxvec(realdp *vec, integer length);
        
        realdp *initDx;
        realdp SMlambda;
        realdp SMtol;
        integer SMiter;
        integer totalSMiter;
        integer SMCGiter;
        integer totalSMCGiter;
        integer dimNorVec;
        
        /*for adaptive RPG*/
        realdp adavalue;
        
        /*Data for ProximalMap*/
		Vector *VecTMP1;
		Vector *VecTMP2;
		Vector *Weight;
		Vector *BLambda;
		Vector *DLambda;
		realdp *Fu, *Fz;
        realdp *SMz, *SMu, *SMd;
    private:
        void testCalACalAadj(Variable *x, Vector *DLambda, realdp *ELambda);
	};
}; /*end of ROPTLIB namespace*/
#endif // end of RPG_H
