
#include "TestSPDMeanLinfty.h"

using namespace ROPTLIB;

void testSPDMeanLinfty(void)
{
	/*Randomly generate a point on the SPD manifold*/
	integer n = 3, num = 10;
	SPDVariable SPDX(n);
	realdp *initialX = SPDX.ObtainWriteEntireData();
	for (integer i = 0; i < n; i++)
	{
		for (integer j = 0; j < n; j++)
		{
			initialX[i + j * n] = 0;
		}
		initialX[i + i * n] = 1;
	}

	// Define the manifold
	SPDManifold Domain(n);
	Domain.SetHasHHR(true); /*set whether the manifold uses the idea in [HGA2015, Section 4.3] or not*/

	realdp *Ls = new realdp[n * n * num + n * n];
    realdp *Xtrue = new realdp[n * n];
	realdp *tmp = Ls + n * n * num;
	integer info;
	for (integer i = 0; i < num; i++)
	{
		for (integer j = 0; j < n * n; j++)
			tmp[j] = genrandnormal();

		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, tmp, &n, &GLOBAL::DZERO, Ls + i * n * n, &n);


		potrf_(GLOBAL::L, &n, Ls + i * n * n, &n, &info);
		if (info != 0)
		{
			std::cout << "Warning: TestSPDMean Cholesky decomposition fails with info:" << info << "!" << std::endl;
		}
		for (integer j = 0; j < n; j++)
			for (integer k = j + 1; k < n; k++)
				Ls[j + k * n + i * n * n] = 0;
	}

	// Define the problem
	SPDMeanLinfty Prob(Ls, n, num);
	/*The domain of the problem is a SPD manifold*/
	Prob.SetDomain(&Domain);

	//Prob.CheckGradHessian(&SPDX);

	/*Output the parameters of the domain manifold*/
	//Domain.CheckParams();

	/*Check the correctness of the manifold operations*/
	//Domain.CheckIntrExtr(&SPDX);
	//Domain.CheckRetraction(&SPDX);
	//Domain.CheckDiffRetraction(&SPDX);
	//Domain.CheckLockingCondition(&SPDX);
	//Domain.CheckcoTangentVector(&SPDX);
	//Domain.CheckIsometryofVectorTransport(&SPDX);
	//Domain.CheckIsometryofInvVectorTransport(&SPDX);
	//Domain.CheckVecTranComposeInverseVecTran(&SPDX);
	//Domain.CheckTranHInvTran(&SPDX);
	//Domain.CheckHaddScaledRank1OPE(&SPDX);

	// test LRBFGS
	RBFGS *RBFGSsolver = new RBFGS(&Prob, &SPDX);
	RBFGSsolver->LineSearch_LS = WOLFELP;
	RBFGSsolver->Stop_Criterion = PSSUBGRAD;
	RBFGSsolver->Debug = FINALRESULT; //ITERRESULT;// 
	RBFGSsolver->Max_Iteration = 2000;
	RBFGSsolver->Tolerance = static_cast<realdp> (1e-6);
	RBFGSsolver->OutputGap = 10;
	RBFGSsolver->Diffx = static_cast<realdp> (1e-5);
	//RBFGSsolver->CheckParams();
	RBFGSsolver->Run();
	if (RBFGSsolver->Getnormgfgf0() < 1e-6)
		printf("SUCCESS!\n");
	else
		printf("FAIL!\n");
	delete RBFGSsolver;
	delete[] Ls;
    delete[] Xtrue;
};



#ifdef MATLAB_MEX_FILE

std::map<integer *, integer> *CheckMemoryDeleted;

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if (nrhs < 4)
	{
		mexErrMsgTxt("The number of arguments should be at least four.\n");
	}
	realdp *Ls, *X, *Xopt;
	integer n, N, HasHHR;
	Ls = mxGetPr(prhs[0]);
	X = mxGetPr(prhs[1]);
	n = mxGetM(prhs[1]);
	const mwSize *ptrdims = mxGetDimensions(prhs[0]);
	if (mxGetNumberOfDimensions(prhs[0]) == 2)
		N = 1;
	else
		N = ptrdims[2];
	HasHHR = static_cast<integer> (mxGetScalar(prhs[2]));

	if (ptrdims[1] != n || ptrdims[0] != n)
	{
		mexErrMsgTxt("The size of matrix C is not correct.\n");
	}
	if (mxGetM(prhs[1]) != n || mxGetN(prhs[1]) != n)
	{
		mexErrMsgTxt("The size of the initial X is not correct!\n");
	}

	genrandseed(0);

	CheckMemoryDeleted = new std::map<integer *, integer>;
	//testStieSoftICA(Cs, n, p, N, X, Xopt);

	// Obtain an initial iterate from the input
	SPDVariable *SPDX = new SPDVariable(n);
	realdp *initialX = SPDX->ObtainWriteEntireData();
	for (integer i = 0; i < n * n; i++)
	{
		initialX[i] = X[i];
	}
    
    
    SPDVariable *SPDsoln = nullptr;
	if (nrhs >= 5)
	{
		realdp *soln = mxGetPr(prhs[4]); /*soln: n by p*/
		SPDsoln = new SPDVariable(n);
		realdp *SPDsolnptr = SPDsoln->ObtainWriteEntireData();
		for (integer i = 0; i < n * n; i++)
		{
			SPDsolnptr[i] = soln[i];
		}
	}

	// Define the manifold
	SPDManifold Domain(n);
	// Define the SPDMean problem
	SPDMeanLinfty Prob(Ls, n, N);
	Prob.SetDomain(&Domain);
	Domain.SetHasHHR((HasHHR != 0));
	ParseSolverParamsAndOptimizing(prhs[3], &Prob, SPDX, SPDsoln, plhs);
    
    
    
    mexProblem::ObtainElementFromMxArray(SPDX, plhs[0]);

	integer numdim = Domain.GetIntrDim() * 10;
	Vector **dgfs = new Vector*[numdim];
	for (integer i = 0; i < numdim; i++)
	dgfs[i] = Domain.GetEMPTYINTR()->ConstructEmpty();

	integer length = 0;
	realdp testzero = 0;
    	plhs[20] = mxCreateDoubleScalar(testzero);
	for (integer i = 0; i < numdim; i++)
		delete dgfs[i];
	delete[] dgfs;

	std::map<integer *, integer>::iterator iter = CheckMemoryDeleted->begin();
	for (iter = CheckMemoryDeleted->begin(); iter != CheckMemoryDeleted->end(); iter++)
	{
		if (iter->second != 1)
			std::cout << "Global address:" << iter->first << ", sharedtimes:" << iter->second << std::endl;
	}
	delete CheckMemoryDeleted;
    delete SPDX;
    delete SPDsoln;
	return;
    

    
// 	//Domain.CheckParams();
// 	std::map<integer *, integer>::iterator iter = CheckMemoryDeleted->begin();
// 	for (iter = CheckMemoryDeleted->begin(); iter != CheckMemoryDeleted->end(); iter++)
// 	{
// 		if (iter->second != 1)
// 			std::cout << "Global address:" << iter->first << ", sharedtimes:" << iter->second << std::endl;
// 	}
// 	delete CheckMemoryDeleted;
// 	delete SPDX;
//  delete SPDsoln;
// 	return;
}

#endif
