    
#include "TestSPDMeanSymmLDOneParamL1.h"

using namespace ROPTLIB;

void testSPDMeanSymmLDOneParamL1(void)
{
	/*Randomly generate a point on the SPD manifold*/
	integer n = 100, num = 4;
    realdp alpha = 0.5;
	SPDVariable SPDX(n);
	realdp *initialX = SPDX.ObtainWriteEntireData();
	for (integer i = 0; i < n; i++)
	{
		for (integer j = 0; j < n; j++)
		{
			initialX[i + j * n] = 0;
		}
		initialX[i + i * n] = 1;
	}

	// Define the manifold
	SPDManifold Domain(n);
	Domain.SetHasHHR(true); /*set whether the manifold uses the idea in [HGA2015, Section 4.3] or not*/

	realdp *Ls = new realdp[n * n * num + n * n];
    realdp *LDAs = new realdp[num];
    realdp *Xtrue = new realdp[n * n];
	realdp *tmp = Ls + n * n * num;
	for (integer i = 0; i < num; i++)
	{
		for (integer j = 0; j < n * n; j++)
			tmp[j] = genrandnormal();

		gemm_(GLOBAL::N, GLOBAL::T, &n, &n, &n, &GLOBAL::DONE, tmp, &n, tmp, &n, &GLOBAL::DZERO, Ls + i * n * n, &n);
    }
    
    for (integer i = 0; i < num; i++) LDAs[i] = 0.0;
    
	// Define the problem
	SPDMeanSymmLDOneParamL1 Prob(Ls, LDAs, n, num, alpha);
	/*The domain of the problem is a SPD manifold*/
	Prob.SetDomain(&Domain);

	//Prob.CheckGradHessian(&SPDX);

	/*Output the parameters of the domain manifold*/
	//Domain.CheckParams();

	/*Check the correctness of the manifold operations*/
	//Domain.CheckIntrExtr(&SPDX);
	//Domain.CheckRetraction(&SPDX);
	//Domain.CheckDiffRetraction(&SPDX);
	//Domain.CheckLockingCondition(&SPDX);
	//Domain.CheckcoTangentVector(&SPDX);
	//Domain.CheckIsometryofVectorTransport(&SPDX);
	//Domain.CheckIsometryofInvVectorTransport(&SPDX);
	//Domain.CheckVecTranComposeInverseVecTran(&SPDX);
	//Domain.CheckTranHInvTran(&SPDX);
	//Domain.CheckHaddScaledRank1OPE(&SPDX);

	// test LRBFGS
	//std::cout << "********************************Test Geometric mean in LRBFGS*************************************" << std::endl;
	LRBFGS *LRBFGSsolver = new LRBFGS(&Prob, &SPDX);
	LRBFGSsolver->LineSearch_LS = ARMIJO;
	LRBFGSsolver->Debug = FINALRESULT; //ITERRESULT;// 
	LRBFGSsolver->Max_Iteration = 20;
	LRBFGSsolver->Tolerance = static_cast<realdp> (1e-6);
	LRBFGSsolver->Accuracy = static_cast<realdp> (1e-6);
	LRBFGSsolver->Finalstepsize = 1;
	//LRBFGSsolver->CheckParams();
	LRBFGSsolver->Run();
	if (LRBFGSsolver->Getnormgfgf0() < 1e-6)
		printf("SUCCESS!\n");
	else
		printf("FAIL!\n");
	delete LRBFGSsolver;

	delete[] Ls;
	delete[] Xtrue;
	delete[] LDAs;
};

#ifdef MATLAB_MEX_FILE

std::map<integer *, integer> *CheckMemoryDeleted;

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if (nrhs < 4)
	{
		mexErrMsgTxt("The number of arguments should be at least four.\n");
	}
	realdp *Ls, *LDAs, *X, *Xopt, *soln;
    realdp alpha;
    
    Ls = mxGetPr(prhs[0]);
    LDAs = mxGetPr(prhs[1]);
    alpha = static_cast<double> (mxGetScalar(prhs[2]));
    X = mxGetPr(prhs[3]);
    
    /* dimensions of input matrices */ 
    integer n, N, HasHHR;
    n = mxGetM(prhs[3]);
    
	const mwSize *ptrdims = mxGetDimensions(prhs[0]);
	if (mxGetNumberOfDimensions(prhs[0]) == 2)
		N = 1;
	else
		N = ptrdims[2];    

	if (ptrdims[1] != n || ptrdims[0] != n)
	{
		mexErrMsgTxt("The size of matrix C is not correct.\n");
	}
	if (mxGetM(prhs[3]) != n || mxGetN(prhs[3]) != n)
	{
		mexErrMsgTxt("The size of the initial X is not correct!\n");
	}
    
    HasHHR = static_cast<integer> (mxGetScalar(prhs[4]));

    // create output matrix
    plhs[0] = mxCreateDoubleMatrix(n, n, mxREAL);
	Xopt = mxGetPr(plhs[0]);

	genrandseed(0);

	CheckMemoryDeleted = new std::map<integer *, integer>;
    
    
	// Obtain an initial iterate from the input
	SPDVariable *SPDX = new SPDVariable(n);
	realdp *initialX = SPDX->ObtainWriteEntireData();
	for (integer i = 0; i < n * n; i++)
	{
		initialX[i] = X[i];
	}
    
    
    SPDVariable *SPDsoln = nullptr;
    if (nrhs >= 7)
	{
		realdp *soln = mxGetPr(prhs[6]);
		SPDsoln = new SPDVariable(n);
		realdp *SPDsolnptr = SPDsoln->ObtainWriteEntireData();
		for (integer i = 0; i < n * n; i++)
		{
			SPDsolnptr[i] = soln[i];
		}
	}


	// Define the manifold
	SPDManifold Domain(n);
	// Define the SPDMean problem
	SPDMeanSymmLDOneParamL1 Prob(Ls, LDAs, n, N, alpha);
	Prob.SetDomain(&Domain);
	Domain.SetHasHHR((HasHHR != 0));
	ParseSolverParamsAndOptimizing(prhs[5], &Prob, SPDX, SPDsoln, plhs);
    

	//Domain.CheckParams();
	std::map<integer *, integer>::iterator iter = CheckMemoryDeleted->begin();
	for (iter = CheckMemoryDeleted->begin(); iter != CheckMemoryDeleted->end(); iter++)
	{
		if (iter->second != 1)
			std::cout << "Global address:" << iter->first << ", sharedtimes:" << iter->second << std::endl;
	}
	delete CheckMemoryDeleted;
	delete SPDX;
    delete SPDsoln;
	return;
}

#endif
